// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class body_sources (private base class of class bodies)                     |
// class body_sinks   (private base class of class bodies)                     |
// class bodies                                                                |
// class body                                                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_body_h
#define included_body_h 1

#ifndef included_auxx_h
#  include <nbdy/auxx.h>
#endif

#ifndef included_inln_h
#  include <util/inln.h>
#endif

#ifndef included_flag_h
#  include <nbdy/flag.h>
#endif

#ifndef included_nbio_h
#  include <nbdy/nbio.h>
#endif

#ifndef included_kern_h
#  include <nbdy/kern.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::body_sources                                                   
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // - allocates source properties of all bodies in one block of memory         
  // - provides access to individual flags, masses, positions, [eps_i]          
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class body_sources {                             // holding the source props  
  private:                                         //   of all bodies           
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
    uint        NBODIES, NALLOC;                   // # allocated, used         
    char       *ALLOC;                             // pointer to memory         
  protected:                                       //                           
    basic_flag *FLAG;                              // flags                     
    real       *MASS;                              // masses                    
    vect       *POS;                               // positions                 
    real       *EPS;                               // softening lengths         
  private:                                         //   of all bodies           
    //--------------------------------------------------------------------------
    // allocate & initialize                                                    
    //--------------------------------------------------------------------------
    void allocate(const uint n  = 0,               // [I: N_bodies]             
		  const bool e  = false,           // [I: allow for eps_i ?]    
		  const uint na = 0) {             // [I: N_alloc > N_bodies]   
      NBODIES = n;                                 //   set N_bodies            
      NALLOC  = na? util::max(n,na) : n;           //   set N_alloc             
      if(NALLOC) {                                 //   IF allocation wanted >  
	ALLOC = e?                                 //     allocate memory       
	  new char[NALLOC*(sizeof(basic_flag)+2*sizeof(real)+sizeof(vect))] :
	  new char[NALLOC*(sizeof(basic_flag)+  sizeof(real)+sizeof(vect))] ;
	FLAG  = (basic_flag*)(ALLOC);
	MASS  = (real*)      (FLAG + NALLOC);
	POS   = (vect*)      (MASS + NALLOC);
	EPS   = e? (real*)   (POS  + NALLOC) : 0;
	for(register basic_flag*Fi=FLAG; Fi!=FLAG+NALLOC; ++Fi) Fi->reset_flag();
      } else {                                     //   > ELSE(no allocation) > 
	ALLOC = 0;                                 //     set pointers = 0      
	FLAG  = 0;
	MASS  = 0;
	POS   = 0;
	EPS   = 0;
      }                                            //   <                       
    }
    //--------------------------------------------------------------------------
    // construction & destruction is protected (to be done by class bodies)     
    //--------------------------------------------------------------------------
  protected:                                       //                           
    body_sources(                                  // constructor               
		 const uint n  = 0,                // [I: N_bodies]             
		 const bool e  = false,            // [I: allow for eps_i ?]    
		 const uint na = 0) {              // [I: N_alloc > N_bodies]   
      allocate(n,e,na);                            //   allocate & initialize   
    }
    //--------------------------------------------------------------------------
    ~body_sources() {                              // destructor                
      if(ALLOC) delete[] ALLOC;                    //   delete memory           
    }
    //--------------------------------------------------------------------------
    void reset  (                                  // destruct & construct      
		 const uint n,                     //  I: N_bodies              
		 const bool e  = false,            // [I: allow for eps_i ?]    
		 const uint na = 0) {              // [I: N_alloc > N_bodies]   
      if(ALLOC) delete[] ALLOC;                    //   delete memory           
      allocate(n,e,na);                            //   allocate & initialize   
    }
    //--------------------------------------------------------------------------
    // const member methods                                                     
    //--------------------------------------------------------------------------
  public:
    const uint &N_bodies()             const { return NBODIES; }
    const uint &N_alloc ()             const { return NALLOC; }
    bool        has_epsi()             const { return EPS!=0; }
    basic_flag &flag    (const uint&i) const { return FLAG[i]; }
    real       &mass    (const uint&i) const { return MASS[i]; }
    vect       &pos     (const uint&i) const { return POS[i]; }
    real       &eps     (const uint&i) const { return EPS[i]; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::body_sinks                                                     
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // - allocates sink properties of all bodies in one block of memory           
  // - provides access to individual pot, acc, rho                              
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class body_sinks {                               // holding the sink props    
  private:                                         //   of all bodies           
    //--------------------------------------------------------------------------
    // data members                                                             
    //--------------------------------------------------------------------------
    uint        NBODIES, NALLOC;                   // # allocated, used         
    char       *ALLOC;                             // pointer to memory         
  protected:                                       //                           
    real       *POT;                               // potentials                
    vect       *ACC;                               // accelerations             
    real       *RHO;                               // densities                 
  private:                                         //   of all bodies           
    //--------------------------------------------------------------------------
    // allocate & initialize                                                    
    //--------------------------------------------------------------------------
    void allocate(const uint n  = 0,               // [I: N_bodies]             
		  const uint na = 0) {             // [I: N_alloc > N_bodies]   
      NBODIES = n;                                 //   set N_bodies            
      NALLOC  = na? util::max(n,na) : n;           //   set N_alloc             
      if(NALLOC) {                                 //   IF allocation wanted >  
	ALLOC = new char[NALLOC*(2*sizeof(real)+sizeof(vect))] ;
	POT   = (real*)(ALLOC);
	ACC   = (vect*)(POT + NALLOC);
	RHO   = (real*)(ACC + NALLOC);
      } else {                                     //   > ELSE(no allocation) > 
	ALLOC = 0;                                 //     set pointers = 0      
	POT   = 0;
	ACC   = 0;
	RHO   = 0;
      }                                            //   <                       
    }
    //--------------------------------------------------------------------------
    // construction & destruction is protected (to be done by class bodies)     
    //--------------------------------------------------------------------------
  protected:                                       //                           
    body_sinks(                                    // constructor               
	       const uint n  = 0,                  // [I: N_bodies]             
	       const uint na = 0) {                // [I: N_alloc > N_bodies]   
      allocate(n,na);                              //   allocate & initialize   
    }
    //--------------------------------------------------------------------------
    ~body_sinks() {                                // destructor                
      if(ALLOC) delete[] ALLOC;                    //   delete memory           
    }
    //--------------------------------------------------------------------------
    void reset(                                    // destruct & construct      
	       const uint n,                       //  I: N_bodies              
	       const uint na = 0) {                // [I: N_alloc > N_bodies]   
      if(ALLOC) delete[] ALLOC;                    //   delete memory           
      allocate(n,na);                              //   allocate & initialize   
    }
    //--------------------------------------------------------------------------
    // const member methods                                                     
    //--------------------------------------------------------------------------
  public:
    const uint &N_bodies()             const { return NBODIES; }
    const uint &N_alloc ()             const { return NALLOC;  }
    vect       &acc     (const uint&i) const { return ACC[i]; }
    real       &pot     (const uint&i) const { return POT[i]; }
    real       &rho     (const uint&i) const { return RHO[i]; }
  };
  //////////////////////////////////////////////////////////////////////////////
#ifdef ALLOW_NEMO
  class nemo_in;                                   // forward declaration       
  class nemo_out;                                  // forward declaration       
#endif
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::bodies                                                         
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // - allocates bodies in one block                                            
  // - provides iterators for sequential access to all bodies                   
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class bodies :
    public body_sources,                           // body source properties    
    public body_sinks    {                         // body sink   properties    
  private:
    bodies(const bodies&);                         // not implemented           
    bodies& operator=(const bodies&);              // not implemented           
    //--------------------------------------------------------------------------
    // data members (apart from those already in body_sources & body_sinks)     
    //--------------------------------------------------------------------------
                                                   // 1 data of all bodies      
    uint                  N, NSPH;                 //    # bodies in total      
                                                   // 2 data of bodies          
                                                   // 2.1 basic properties      
    vect                 *VEL;                     //     velocities            
                                                   // 2.2 support for blocksteps
    mutable indx         *LEVEL;                   //     time-step level       
                                                   // 2.3 support for SPH       
    real                 *SIZE;                    //     SPH size of body      
                                                   // 2.4 support for P_ext     
    mutable real         *PEX;                     //     external potential    
                                                   // 2.5 support for aux       
    mutable real         *AUX;                     //     auxiliary data        
    //--------------------------------------------------------------------------
    // construction & destruction                                               
    //--------------------------------------------------------------------------
  public:
    bodies() :
      body_sources(0), body_sinks(0),
      N(0), NSPH(0), VEL(0), 
      LEVEL(0),
      SIZE(0), PEX(0), AUX(0)
      {}
    //--------------------------------------------------------------------------
    bodies(                                        // constructor               
	   const uint n,                           // I: # bodies               
	   const bool e    = false,                //[I: add indiv soft lengths]
	   const uint nsph = 0u,                   //[I: first N_sph: SPH]      
	   const bool s    = false,                //[I: add stepping support]  
	   const bool p    = false,                //[I: add P_ext support]     
	   const bool a    = false) :              //[I: add aux support]       
      body_sources( n, e),
      body_sinks  ( n ),
      N           ( n ),
      NSPH        ( util::min(n,nsph) ),
      VEL         ( N?        new vect[N]    : 0 ),
      LEVEL       ( (s && N)? new indx[N]    : 0 ),
      SIZE        ( NSPH?     new real[NSPH] : 0 ),
      PEX         ( (p && N)? new real[N]    : 0 ),
      AUX         ( (a && N)? new real[N]    : 0 )
    {
      if(NSPH)
	for(register uint i=0; i!=NSPH; ++i)
	  flag(i).add_flag(basic_flag::SPH);
    }
    //--------------------------------------------------------------------------
#define RESET(NAME,TYPE,NUMBER) 		\
    if(NAME) delete[] NAME;			\
    NAME = NUMBER? new TYPE[NUMBER] : 0;
    void reset(                                    // change N & NSPH           
	       const uint n,                       // I: # bodies               
	       const bool e    = false,            //[I: add indiv soft lengths]
	       const uint nsph = 0u)               //[I: first N_sph: SPH]      
    {
      if(nsph != NSPH) {
	NSPH = util::min(nsph,n);
	RESET(SIZE, real,NSPH)
      }
      if(n!=N || e!=has_epsi() ) body_sources::reset(n,e);
      if(n!=N)                   body_sinks::reset(n);
      if(n != N) {
	N = n;
	RESET(VEL,  vect,N)
#undef RESET
#define RESET(NAME,TYPE,NUMBER)				\
    if(NAME) { delete[] NAME;				\
	       NAME = NUMBER? new TYPE[NUMBER] : 0; }
	RESET(LEVEL,indx,N)
	RESET(PEX,  real,N)
	RESET(AUX,  real,N)
#undef RESET
      }
      if(NSPH)
	for(register uint i=0; i!=NSPH; ++i)
	  flag(i).add_flag(basic_flag::SPH);
    }
    //--------------------------------------------------------------------------
    ~bodies() {
      if(VEL)   delete[] VEL;
      if(LEVEL) delete[] LEVEL;
      if(SIZE)  delete[] SIZE;
      if(PEX)   delete[] PEX;
      if(AUX)   delete[] AUX;
    }
    //--------------------------------------------------------------------------
    // I/O  both nemo and yanc format                                           
    //--------------------------------------------------------------------------
#ifdef ALLOW_NEMO
    //--------------------------------------------------------------------------
    io read_nemo_snapshot(                             // R: what has been read 
			  const nemo_in&,              // I: nemo input         
			  real*         =0,            //[I: read time]         
			  const io      =io::mxv,      //[I: what to read]      
			  char*         =0,            //[I: time range]        
			  const bool    =true);        //[I: warn: missing data]
    //--------------------------------------------------------------------------
    io read_nemo_particles(                            // R: what has been read 
			   const nemo_in&,             // I: nemo input         
			   real*         =0,           //[I: read time]         
			   const io      =io::mxv,     //[I: what to read]      
			   char*         =0,           //[I: time range]        
			   const bool    =true);       //[I: warn: missing data]
    //--------------------------------------------------------------------------
    void write_nemo_snapshot(                          // write snapshot        
			     nemo_out&,                // I: nemo output        
			     const real* =0,           //[I: write time]        
			     const io =io::mxv) const; //[I: what to write]     
    //--------------------------------------------------------------------------
    void write_nemo_particles(                         // write bodies to output
			      nemo_out&,               // I: nemo output        
			      const real* =0,          //[I: write time]        
			      const io =io::mxv) const;//[I: what to write]     
#endif
    //--------------------------------------------------------------------------
    void write_yanc_ascii(                             // write ascii to ostream
			  std::ostream&,               // I: output stream      
			  const io=io::mxv) const;     //[I: what to write out ]
    //--------------------------------------------------------------------------
    io   read_yanc_ascii(                              // R: what has been read 
			 std::istream&,                // I: input stream       
			 const io=io::mxv);            //[I: what to read in]   
    //--------------------------------------------------------------------------
    void write_yanc_binary(                            // write bin to ostream  
			   std::ostream&,              // I: output stream      
			   const io =io::mxv) const;   //[I: what to write]     
    //--------------------------------------------------------------------------
    void read_yanc_binary (                            // read bin from istream 
			   std::istream&,              // I: input stream       
			   const io =io::mxv);         //[I: what to read]      
    //--------------------------------------------------------------------------
    // direct data access                                                       
    //                                                                          
    // to be used instead of the arrays themselves                              
    // (this is to simplify possible changes in memory layout: they only affect 
    // these few inline functions)                                              
    //--------------------------------------------------------------------------
    vect &vel   (const uint&i) const { return VEL[i]; }
    indx &level (const uint&i) const { return LEVEL[i]; }
    real &size  (const uint&i) const { return SIZE[i]; }
    real &pex   (const uint&i) const { return PEX[i]; }
    real &aux   (const uint&i) const { return AUX[i]; }
    //--------------------------------------------------------------------------
    // allocation & de-allocation                                               
    //--------------------------------------------------------------------------
    void enable_time_steps() const {
      if(N) {
	if(LEVEL==0) LEVEL = new indx[N];
      }
    }
    //--------------------------------------------------------------------------
    void disable_time_steps() const {
      if(LEVEL) { delete[] LEVEL; LEVEL=0; }
    }
    //--------------------------------------------------------------------------
    void enable_external_pot() const {
      if(PEX==0 && N!=0) PEX = new real[N];
    }
    //--------------------------------------------------------------------------
    void disable_external_pot() const {
      if(PEX) { delete[] PEX; PEX=0; }
    }
    //--------------------------------------------------------------------------
    void enable_aux_data() const {
      if(AUX==0 && N!=0) AUX = new real[N];
    }
    //--------------------------------------------------------------------------
    void disable_aux_data() const {
      if(AUX) { delete[] AUX; AUX=0; }
    }
    //--------------------------------------------------------------------------
    // other non-constant methods                                               
    //--------------------------------------------------------------------------
    void flag_all_as_sink();                       // flag all bodies as sinks  
    //--------------------------------------------------------------------------
    // constant methods                                                         
    //--------------------------------------------------------------------------
    const uint& N_body            () const { return N; }
    const uint& N_sources         () const { return body_sources::N_alloc(); }
    const uint& N_sinks           () const { return body_sinks::N_alloc(); }
    const uint& N_SPH             () const { return NSPH; }
    bool        support_timesteps () const { return LEVEL; }
    bool        support_SPH       () const { return NSPH; }
    bool        support_Pex       () const { return PEX; }
    bool        support_aux       () const { return AUX; }
    //--------------------------------------------------------------------------
    // public types                                                             
    //--------------------------------------------------------------------------
  public:
    class iterator {
      //........................................................................
      // basic stuff (construction, iteration, etc)                             
      //........................................................................
      friend class bodies;
    private:
      iterator();                                  // not implemented           
      //........................................................................
      // private types                                                          
      //........................................................................
      enum {
	MARK_BODY   = 1<<9,                        // use  9th bit of basic_flag
	NOT_LONGER  = 1<<10,                       // use 10th bit of basic_flag
	NOT_SHORTER = 1<<11                        // use 11th bit of basic_flag
      };
    protected:
      //........................................................................
      // data members                                                           
      //........................................................................
      const bodies *B;                             // pointer to OBJECTS        
      unsigned      K;                             // index                     
      //........................................................................
      // construction                                                           
      //........................................................................
      iterator(const bodies*b, unsigned k) : B(b), K(k) {}
    public:
      iterator(const iterator&I)           : B(I.B), K(I.K) {}
      //........................................................................
      // forward iteration                                                      
      //........................................................................
      iterator& operator= (const iterator&I)       { B=I.B; K=I.K; return *this; }
      iterator& operator++()                       { ++K; return *this; }
      iterator  operator++(int)                    { return iterator(B,K++); }
      iterator& operator+=(const unsigned&k)       { K+=k; return *this; }
      iterator  operator+ (const unsigned&k) const { return iterator(B,K+k); }
      //........................................................................
      // boolean methods                                                        
      //........................................................................
      bool  operator== (const iterator&I) const { return K == I.K; }
      bool  operator!= (const iterator&I) const { return K != I.K; }
      bool  operator<  (const iterator&I) const { return K <  I.K; }
      bool  operator<= (const iterator&I) const { return K <= I.K; }
      bool  operator>  (const iterator&I) const { return K >  I.K; }
      bool  operator>= (const iterator&I) const { return K >= I.K; }
      //........................................................................
      // data access                                                            
      //........................................................................
#define DATA_ACCESS_REF(TYPE,NAME)						\
             TYPE      &NAME()                 { return B->NAME(K); }		\
             TYPE const&NAME()           const { return B->NAME(K); }		\
      friend TYPE const&NAME(const iterator&I) { return I.NAME(); }
      //........................................................................
  public:
      iterator               next   ()           const { return iterator(B,K+1); }
      friend unsigned  const&index  (const iterator&I) { return I.K; }
      DATA_ACCESS_REF(real,mass)
      DATA_ACCESS_REF(vect,pos)
      real&pos  (const int i) { return pos()[i]; }
      DATA_ACCESS_REF(vect,vel)
      real&vel  (const int i) { return vel()[i]; }
      DATA_ACCESS_REF(vect,acc)
      real&acc  (const int i) { return acc()[i]; }
      DATA_ACCESS_REF(real,eps)
      DATA_ACCESS_REF(real,pot)
      DATA_ACCESS_REF(real,pex)
      DATA_ACCESS_REF(real,aux)
      DATA_ACCESS_REF(real,rho)
      DATA_ACCESS_REF(indx,level)
      DATA_ACCESS_REF(real,size)
      DATA_ACCESS_REF(basic_flag,flag)
#undef DATA_ACCESS_REF
      //........................................................................
      // flag manipulations                                                     
      //........................................................................
      bool flag_is_set(const flag_type&F) const { return flag().is_set(F); }
      operator const basic_flag& ()       const { return flag(); }
      void flag_as_sink      () { flag().add_flag(basic_flag::SINK); }
      void unflag_sink       () { flag().un_flag (basic_flag::SINK); }
      void flag_as_sticky    () { flag().add_flag(basic_flag::STICKY); }
      void unflag_sticky     () { flag().un_flag (basic_flag::STICKY); }
      void flag_as_sph       () { flag().add_flag(basic_flag::SPH); }
      void unflag_sph        () { flag().un_flag (basic_flag::SPH); }
      void allow_tree_usage  () { flag().un_flag (basic_flag::NOT_IN_TREE); }
      void forbid_tree_usage () { flag().add_flag(basic_flag::NOT_IN_TREE); }
      void allow_longer      () { flag().un_flag (NOT_LONGER); }
      void forbid_longer     () { flag().add_flag(NOT_LONGER); }
      void allow_shorter     () { flag().un_flag (NOT_SHORTER); }
      void forbid_shorter    () { flag().add_flag(NOT_SHORTER); }
      void mark              () { flag().add_flag(MARK_BODY); }
      void unmark            () { flag().un_flag (MARK_BODY); }
      //........................................................................
      // const boolean informations via mambers                                 
      //........................................................................
      bool is_source     () const { return mass() != zero; }
      bool may_go_longer () const { return !flag_is_set(NOT_LONGER);}
      bool may_go_shorter() const { return !flag_is_set(NOT_SHORTER); }
      bool is_marked     () const { return  flag_is_set(MARK_BODY); }
      //........................................................................
      // const boolean informations via friends                                 
      //........................................................................
      friend bool is_source     (const iterator&I) { return I.is_source(); }
      friend bool may_go_longer (const iterator&I) { return I.may_go_longer(); }
      friend bool may_go_shorter(const iterator&I) { return I.may_go_shorter(); }
      friend bool is_marked     (const iterator&I) { return I.is_marked(); }
      //........................................................................
      // output: give just the index (an iterator acts like a pointer)          
      //........................................................................
      friend std::ostream& operator<<(std::ostream&o, const iterator&I) {
	return o<<index(I);
      }
    };
    //--------------------------------------------------------------------------
    typedef class iterator     body_type;          // type of body              
    typedef class body_sources sources_type;       // type of body_sources      
    typedef class body_sinks   sinks_type;         // type of body_sinks        
    //--------------------------------------------------------------------------
    // initializing iterators                                                   
    //--------------------------------------------------------------------------
    iterator body_no     (const uint&i) const { return iterator(this,i); }
    iterator begin_bodies()             const { return iterator(this,0); }
    iterator end_bodies  ()             const { return iterator(this,N); }
    iterator back_bodies ()             const { return iterator(this,N-1); }
    iterator begin_SPH   ()             const { return begin_bodies(); }
    iterator end_SPH     ()             const { return iterator(this,NSPH); }
    iterator back_SPH    ()             const { return iterator(this,NSPH-1); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // typedefs                                                                   
  //////////////////////////////////////////////////////////////////////////////
  typedef bodies::body_type body;                  // define nbdy::body         
}                                                  // namespace nbdy            
////////////////////////////////////////////////////////////////////////////////
// useful macros                                                                
////////////////////////////////////////////////////////////////////////////////
#ifndef  LoopBodies
#define LoopBodies(BODIES_TYPE,                    /* type of bodies         */	\
		   BODIES_PTER,                    /* pointer to bodies      */	\
		   NAME)                           /* name for body          */	\
  for(register BODIES_TYPE::body_type              /* type of body           */	\
      NAME  = BODIES_PTER->begin_bodies();         /* from first body        */	\
      NAME != BODIES_PTER->end_bodies();           /* until beyond last body */	\
    ++NAME)                                        /* get next body          */
#endif
//------------------------------------------------------------------------------
#ifndef  LoopSPHBodies
#define LoopSPHBodies(BODIES_TYPE,                 /* type of bodies         */	\
		      BODIES_PTER,                 /* pointer to bodies      */	\
		      NAME)                        /* name for body          */	\
  for(register BODIES_TYPE::body_type              /* type of body           */	\
      NAME  = BODIES_PTER->begin_SPH();            /* from first body        */	\
      NAME != BODIES_PTER->end_SPH();              /* until beyond last body */	\
    ++NAME)                                        /* get next body          */
#endif
////////////////////////////////////////////////////////////////////////////////
#endif // included_body_h
