// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// auxx.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// some global macros                                                          |
// some global constants                                                       |
// some global typedefs                                                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_auxx_h
#define included_auxx_h

#ifndef included_iostream
#  include <iostream>
#  define included_iostream
#endif
#ifndef included_cmath
#  include <cmath>
#  define included_cmath
#endif
#ifndef included_cstdlib
#  include <cstdlib>
#  define included_cstdlib
#endif

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 1. INTERNAL PRECISION                                                        
//                                                                              
////////////////////////////////////////////////////////////////////////////////

#if defined(DOUBLE_SINGLE) || defined(DOUBLE_DOUBLE)
//                                                                              
// 1.1 double precision: 64bit floating point numbers to hold a real            
//                                                                              
namespace nbdy {
  typedef double real;
  inline real SQRT(const real x) { return std::sqrt(x); }
  inline real EXP (const real x) { return std::exp (x); }
  inline real LOG (const real x) { return std::log (x); }
  inline real POW (const real x,
		   const real y) { return std::pow(x,y); }
}
#else 
//                                                                              
// 1.2 single precision: 32bit floating point numbers to hold a real            
//                                                                              
namespace nbdy {
  typedef float real;
#ifdef linux
  inline real SQRT(const real x) { return sqrtf(x); }
  inline real EXP (const real x) { return expf (x); }
  inline real LOG (const real x) { return logf (x); }
#else
  inline real SQRT(const real x) { return std::sqrt(double(x)); }
  inline real EXP (const real x) { return std::exp (double(x)); }
  inline real LOG (const real x) { return std::log (double(x)); }
#endif
  inline real POW (const real x,
		   const real y) { return std::pow(double(x),double(y)); }
}
#endif
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 2. DATA TRANSFER VIA BODIES OR ARRAYS & PRECISION USED IN ARRAY ARGS         
//                                                                              
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#if defined(SINGLE_DOUBLE) || defined(DOUBLE_DOUBLE)
  typedef double areal;
#else
  typedef float areal;
#endif
}
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 3. DIMENSIONALITY                                                            
//                                                                              
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 3.1 Two dimensions                                                           
//                                                                              
#ifdef TWODIMENSIONAL

#define NDIM 2   // # dimensions
#define NSUB 4   // 2^NDIM

#else // TWODIMENSIONAL
//                                                                              
// 3.2 Three dimensions                                                         
//                                                                              

#define NDIM 3   // # dimensions
#define NSUB 8   // 2^NDIM

#endif // TWODIMENSIONAL

#ifndef included_vect_h
#  include <nbdy/vect.h>
#endif

#ifndef included_tens_h
#  include <nbdy/tens.h>
#endif

namespace nbdy {
  typedef class vector<real>  vect; // plain NDIM dimensional vector       
  typedef class sym1  <real>  ten1; // symmetric tensor of order 1 = vector
  typedef class sym2  <real>  ten2; // symmetric tensor of order 2 = matrix
  typedef class sym3  <real>  ten3; // symmetric tensor of order 3         
  typedef class sym4  <real>  ten4; // symmetric tensor of order 4         
}

#ifdef TWODIMENSIONAL
namespace nbdy {
  typedef real                   amom;
}
#else // TWODIMENSIONAL
namespace nbdy {
  typedef vect                   amom;
}
#endif // TWODIMENSIONAL

////////////////////////////////////////////////////////////////////////////////
//                                                                              
// 4. USEFUL CONSTANTS AND TYPEDEFS                                             
//                                                                              
////////////////////////////////////////////////////////////////////////////////

namespace nbdy {
  typedef unsigned short   indx;         // use only in non-register memory
  typedef unsigned int     uint;
  typedef int              flag_type;
  //----------------------------------------------------------------------------
  inline void error  (const char*m) {
    std::cerr<<" TREE ERROR: "  <<m<<std::endl;
    std::exit(1);
  }
  //----------------------------------------------------------------------------
  inline void warning(const char*m) {
    std::cerr<<" TREE WARNING: "<<m<<std::endl;
  }
  //----------------------------------------------------------------------------
  const   real zero    = 0.,
               sixth   = 0.166666666666666666666667,
               fifth   = 0.2,
               quarter = 0.25,
               third   = 0.333333333333333333333333,
               half    = 0.5,
               one     = 1.,
               two     = 2.,
               three   = 3.,
               four    = 4.,
               six     = 6.,
               ten     = 10.,
               if2     = 0.5,                           // 1 / 2!              
               if3     = 0.166666666666666666666667,    // 1 / 3!              
               if4     = 0.0416666666666666666666667,   // 1 / 4!              
               if5     = 0.00833333333333333333333333,  // 1 / 5!              
               if6     = 0.00138888888888888888888889;  // 1 / 6!              

}
////////////////////////////////////////////////////////////////////////////////
#endif  // included_auxx_h
