// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// soul.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class soul                fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_soul_h
#define included_soul_h

#ifndef included_auxx_h
#  include <nbdy/auxx.h>
#endif

#ifndef included_flag_h
#  include <nbdy/flag.h>
#endif

#ifndef included_bsoul_h
#  include <nbdy/bsoul.h>
#endif

#ifndef included_body_h
#  include <body.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::soul                                                           
  //////////////////////////////////////////////////////////////////////////////
  class soul : public basic_soul<body_sources> {
    soul           (const soul&);                  // not implemented           
    soul& operator=(const soul&);                  // not implemented           
    //--------------------------------------------------------------------------
  public:
    typedef bodies     bodies_type;
    typedef body_sinks sinks_type;
    //--------------------------------------------------------------------------
    // data of class soul                                                       
    //--------------------------------------------------------------------------
  private:
    scalar_type    MASS;                           // mass of body              
    union {
      scalar_type  POT;                            // potential                 
      scalar_type  RHO;                            // mass density              
    };
    vector_type    ACC;                            // acceleration              
    scalar_type    EPS;                            // individual softening      
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    soul() {}
    //--------------------------------------------------------------------------
    // non-const data access via members                                        
    //--------------------------------------------------------------------------
    vector_type&acc() { return ACC; }
    scalar_type&pot() { return POT; }
    scalar_type&rho() { return RHO; }
    scalar_type&eps() { return EPS; }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
    friend vector_type const&cofm  (const soul*const&S) { return S->cofm(); }
    friend unsigned    const&mybody(const soul*const&S) { return S->mybody(); }
    friend vector_type const&acc   (const soul*const&S) { return S->ACC; }
    friend scalar_type const&pot   (const soul*const&S) { return S->POT; }
    friend scalar_type const&rho   (const soul*const&S) { return S->RHO; }
    friend scalar_type const&eps   (const soul*const&S) { return S->EPS; }
    friend scalar_type const&mass  (const soul*const&S) { return S->MASS; }
    //--------------------------------------------------------------------------
    // simple manipulations for use with bodies                                 
    //--------------------------------------------------------------------------
    void soul::set_up(const body_sources*B) {
      MASS   = B->mass(mybody());
      cofm() = B->pos (mybody());
    }
    //--------------------------------------------------------------------------
    void link_and_set_up(const body_sources*B, const uint&i) {
      link  (i);
      set_up(B);
    }
    //-------------------------------------------------------------------------
    void reset_grav(const body_sources*B) {
      copy_sink_flag(B);
      if(B->has_epsi()) EPS = B->eps(mybody());
      POT = zero;
      ACC = zero;
    }
    //--------------------------------------------------------------------------
    void update_grav(const body_sinks*B) const {
      B->pot(mybody()) = POT;
      B->acc(mybody()) = ACC;
    }
    //--------------------------------------------------------------------------
    void update_dens(const body_sinks*B) const {
      B->rho(mybody()) = RHO;
    }
    //--------------------------------------------------------------------------
    // simple manipulations for use with arrays                                 
    //--------------------------------------------------------------------------
    void set_up(const areal*m, const areal*x[NDIM]) {
      MASS = m[mybody()];
      for(register int d=0; d<NDIM; ++d) cofm()[d] = x[d][mybody()];
    }
    //--------------------------------------------------------------------------
    void link_and_set_up(const vector_type x, const uint i) {
      link(i);
      cofm()=x;
    }
    //--------------------------------------------------------------------------
    void set_mass(const areal*m) {
      MASS = m[mybody()];
    }
    //--------------------------------------------------------------------------
    void update_grav(areal*a[NDIM], areal*p) const {
      if(p) p[mybody()] = POT;
      for(register int d=0; d<NDIM; ++d) a[d][mybody()] = ACC[d];
    }
    //--------------------------------------------------------------------------
    void update_dens(areal*rh) const {
      rh[mybody()] = RHO;
    }
    //--------------------------------------------------------------------------
    void reset_grav(const flag_type*F) { 
      copy_sink_flag(F);
      POT = zero;
      ACC = zero;
    }
    //--------------------------------------------------------------------------
    void reset_grav(const flag_type*F, const areal* E) { 
      copy_sink_flag(F);
      EPS = E[mybody()];
      POT = zero;
      ACC = zero;
    }
    //--------------------------------------------------------------------------
    void reset_grav_mass(const flag_type*F,
			 const areal* M) { 
      copy_sink_flag(F);
      MASS = M[mybody()];
      POT  = zero;
      ACC  = zero;
    }
    //--------------------------------------------------------------------------
    void reset_grav_mass(const flag_type*F,
			 const areal* M, const areal* E) { 
      copy_sink_flag(F);
      MASS = M[mybody()];
      EPS  = E[mybody()];
      POT  = zero;
      ACC  = zero;
    }
    //--------------------------------------------------------------------------
    // simple manipulations for general use                                     
    //--------------------------------------------------------------------------
    void normalize_grav () {                       // acc,pot     /= mass       
      if(MASS>zero) {
	register scalar_type im = one/MASS;
	POT *= im;
	ACC *= im;
      }
    }
    //--------------------------------------------------------------------------
    // boolean information                                                      
    //--------------------------------------------------------------------------
    friend bool is_source(const soul*const&S) { return mass(S) != zero; }
    //--------------------------------------------------------------------------
    // dump soul data                                                           
    //--------------------------------------------------------------------------
    static void dump_head(ostream& o) {
      o<<"     #";
      basic_soul<body_sources>::dump_head(o);
      o<<"              mass";
    }
    //----------------------------------------------------------------------------
    void dump(ostream&o, const soul*S0) const
    {
      o<<" "<<setw(5)<<int(this-S0);
      basic_soul<body_sources>::dump(o);
      o<<" "<<setw(8)<<MASS;
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_soul_h
