////////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <ctime>
#include <cstdlib>
#ifdef unix
#  define GIVE_HOST_INFO
  extern "C" {
#  include <unistd.h>
#  include <pwd.h>
  }
#endif
#include <body.h>
#include <util/Pi.h>
#include <util/rand.h>

using namespace nbdy;
using namespace util;
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  
  const double Msun   = 4.498658966346282e-6; // solar mass in N-body units
  const double kms    = 1./0.977775320024919; // km/s in N-body units
  
  if(argc!=6 && argc!=12) {
    cerr<<
      " \"Plummer N S M R Sym [x,y,z, vx,vy,vz]\" with \n"
      "  N:                number of positions & velocities\n"
      "  S = long:         seed for RNG if S=0, use sobol\n"
      "  M:                if > 0: mass in solar units\n"
      "                    if = 0: use natural rather than N-body units\n"
      "  R:                scale radius in kpc (N-body units only)\n"
      "  C = 0/1:          ensure centroid?\n"
      "  [x,y,z,vx,vy,vz]: optional: position of origin [kpc,km/s]\n"
      " units imply G=1 and\n"
      "  either imply M=1, Rc/Rt=1 (if M=0 given)\n"
      "  or     are kpc and Myr (if M is given in M_sun)\n";
    exit(1);
  }

  const double   twothird=2/3.;
  bool           shift=false,natural=false;
  int            p=1, N, Sy;
  double         Mt,mass,rscl,vscl;
  vector<double> X0=0., V0=0.;
  long           S;
  //----------------------------------------------------------------------------
  // 1. read parameters                                                         
  //----------------------------------------------------------------------------
  N    = atoi(argv[p++]);
  S    = atoi(argv[p++]);
  Mt   = atof(argv[p++]);
  rscl = atof(argv[p++]);
  Sy   = atoi(argv[p++]);
  if(Mt>0.) {
    mass = Mt*Msun;
    vscl = sqrt(mass/rscl);
  } else {
    natural = true;
    mass = 1.;
    rscl = 1.;
    vscl = 1.;
  }
  if(argc==13) {
    shift = true;
    for(register int i=0; i!=3; ++i) X0[i] = atof(argv[p++]);
    for(register int i=0; i!=3; ++i) V0[i] = atof(argv[p++]);
    if(!natural) V0 *= kms;
  }
  //----------------------------------------------------------------------------
  // 2. write header                                                            
  //----------------------------------------------------------------------------
#ifdef GIVE_HOST_INFO
  time_t now = time(0);
  char host[100]; gethostname(host,100);
#endif
  cout<<
    "## -----------------------------------------------------------------------\n"
    "## Plummer model initial conditions in YANC format\n"
    "## created\n"
#ifdef GIVE_HOST_INFO
    "## on              "<<ctime(&now)<<
    "## with program    \"Plummer\" of Walter Dehnen (MPIA Heidelberg)\n"
    "## by user         \""<<(getpwuid(geteuid())->pw_name)<<"\"\n"
    "## on machine      \""<<host<<"\"\n##\n";
#else
    "## with program    \"Plummer\" of Walter Dehnen (MPIA Heidelberg)\n##\n";
#endif
  if(natural)
    cout<<"## in natural units (G=M=Rs=1)\n";
  else {
    cout<<
      "## in N-body units (G=1,[R]=kpc,[T]=Myr)\n"
      "## total mass:     "<<mass<<" = "<<Mt<<" Msun\n"
      "## core  radius:   "<<rscl<<" kpc\n";
  }
  if(shift) cout<<"## shifted to pos: "<<X0<<"\n"
		<<"## shifted to vel: "<<V0<<"\n";
  if(S) cout<<"## using pseudo-random numbers\n";
  else  cout<<"## using quasi-random numbers\n";
  cout<<
    "## -----------------------------------------------------------------------\n"
    "# time           0\n"
    "# N              "<<N<<"\n"
    "# data_format    0\n"
    "# start_data\n";
  //----------------------------------------------------------------------------
  // 3. create initial (x,v)                                                    
  //----------------------------------------------------------------------------
  bodies BB(N);
  const    double m = mass/double(N);
  register double Mr,rq,veq,rad,P,vq,f0,fs,vel,cth,R,phi;
  Sobol           s1,s2,s3,s4,s5,s6,se;
  Random3         R3(S);
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    //                                                                          
    // 3.1 set mass and get r and v                                             
    //                                                                          
    Bi.mass() = m;
    Mr  = pow((S? R3():s1()),twothird);
    rq  = Mr/(1.-Mr);
    rad = sqrt(rq);
    veq = 2./sqrt(rq+1.);
    do {
      P   = veq*pow((S? R3():s2() ),1./4.5);
      vq  = veq-P;
      vel = sqrt(vq);
      f0  = vel*pow(P,3.5);
      fs  = veq*f0;
      f0 *= vel;
    } while (fs* (S? R3() : se())>f0);
    rad*= rscl;
    vel*= vscl;
    //                                                                          
    // 3.2 get x,y,z                                                            
    //                                                                          
    cth       = S? 2*R3()-1 : 2*s2()-1;
    R         = rad * sqrt(1.-cth*cth);
    phi       = TPi*( S? R3() : s3() );
    Bi.pos(0) = R * cos(phi);
    Bi.pos(1) = R * sin(phi);
    Bi.pos(2) = rad * cth;
    //                                                                          
    // 3.3 get vx,vy,vz                                                         
    //                                                                          
    cth       = S? 2*R3()-1 : 2*s5()-1;
    R         = vel * sqrt(1.-cth*cth);
    phi       = TPi*( S? R3() : s6() );
    Bi.vel(0) = R * cos(phi);
    Bi.vel(1) = R * sin(phi);
    Bi.vel(2) = vel * cth;
  }
  //----------------------------------------------------------------------------
  // 4. centrate position and velocity on origin or (x0,v0)                     
  //----------------------------------------------------------------------------
  register vector<double> cpos(0.), cvel(0.);
  if(Sy) {
    LoopBodies(nbdy::bodies,(&BB),Bi) {
      cpos += nbdy::pos(Bi);
      cvel += nbdy::vel(Bi);
    }
    cpos/= double(N);
    cvel/= double(N);
  }
  cpos-= X0;
  cvel-= V0;
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    Bi.pos() -= cpos;
    Bi.vel() -= cvel;
  }
  //----------------------------------------------------------------------------
  // 5. output of (x,v)                                                         
  //----------------------------------------------------------------------------
  BB.write_yanc_ascii(cout,io::mxv);
  cout.flush();
}
