//-----------------------------------------------------------------------------+
//                                                                             |
// falcON.cc                                                                   |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines all function members of class tree                                  |
//                                                                             |
// Contents                                                                    |
//                                                                             |
// 1  Constructor, destructor and such                                         |
// 2  tree growth and re-use                                                   |
// 3  information and statistics                                               |
// 4  estimation of density and such                                           |
// 5  approximation of gravity                                                 |
// 6  neighbour finding and counting                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <falcON.h>                        // class falcON                      
#include <body.h>                          // class body and all its members    
#include <nbdy/soul.h>                     // class soul and all its members    
#include <nbdy/cell.h>                     // class cell and all its members    
#include <nbdy/grat.h>                     // class gravity_tree                
#include <nbdy/stic.h>                     // support for SPH & sticky particles
#include <nbdy/tayl.h>                     // Taylor series & stuff             
#include <nbdy/open.h>                     // class opening                     
#include <nbdy/grav.h>                     // class interactor                  
#include <nbdy/iact.h>                     // generic interaction algorithm     
#include <nbdy/stat.h>                     // statistics & error messages       
#include <iomanip>                         // C++ I/O formatting etc            
#include <fstream>                         // C++ file I/O                      

#ifdef TEST_CPU
#include <ctime>                           // CPU timing                        
using std::clock;                          //                                   
using std::clock_t;                        //                                   
#endif
using std::ostream;                        // we'll use ostream                 
using std::ofstream;                       // and ofstream                      
namespace nbdy {
  typedef gravity_tree tree_type;          // the tree type used                
}
using namespace nbdy;                      // make all tree stuff available     
//=============================================================================+
//                                                                             |
// 1 CONSTRUCTORS, DESTRUCTOR AND SUCH                                         |
// ===================================                                         |
//                                                                             |
//=============================================================================+
falcON::falcON(const bodies*   b,
	       const real      e,
	       const real      th,
	       const kern_type t,
	       const uint      d) :
  STATE(unset),
  BODIES(b), TREE(0),
  Nb(BODIES->N_body()), Ncrit(0), Ncoeffs(0),
  FL(0), M(0), EP(0), X(0), PH(0), RH(0), A(0),
  OPENING(new opening(util::abs(th),(NDIM==2 || th<zero))),
  STATS  (new TreeStats()),
  DMAX   (d)
{
  reset_softening(e,t);
}
//------------------------------------------------------------------------------
falcON::falcON(const int      *f,
	       const areal    *m,
	       const areal    *x[NDIM],
	       const areal    *ep,
	             areal    *a[NDIM],
	             areal    *ph,
	             areal    *rh,
	       const unsigned  n,
	       const real      e,
	       const real      th,
	       const kern_type t,
	       const unsigned  d) :
  STATE(unset),
  BODIES(0), TREE(0),
  Nb(n), Ncrit(0),
  FL(f),M(m),EP(ep),X(x),PH(ph),RH(rh),A(a),
  OPENING(new opening(util::abs(th),(NDIM==2 || th<zero))),
  STATS(new TreeStats()),
  DMAX(d)
{
  reset_softening(e,t);
}
//------------------------------------------------------------------------------
falcON::~falcON() {
  if(TREE) delete TREE;
  delete OPENING;
  delete STATS;
}
//------------------------------------------------------------------------------
void falcON::reset_softening (const real e, const kern_type t) const {
  SOFTENING = (e<zero)? individual : global;
  EPS       = (e<zero)? zero : e;
  KERNEL    = t;
}
//------------------------------------------------------------------------------
void falcON::reset_opening(const real t) const {
  OPENING->reset(util::abs(t),t<zero);
}
//------------------------------------------------------------------------------
inline cell* falcON::root() {
  return TREE->root();
}
//------------------------------------------------------------------------------
inline const cell* falcON::root() const {
  return TREE->root();
}
//=============================================================================+
//                                                                             |
// 2 TREE GROWTH AND RE-USE                                                    |
// ========================                                                    |
//                                                                             |
//=============================================================================+
//------------------------------------------------------------------------------
void falcON::grow(const uint n, const vect*pc)
  // This routines grows a fresh tree & evaluates the cells' basic source       
  // properties.                                                                
{
  Ncrit = util::max(n,1u);
  if(TREE) delete TREE;
  TREE = BODIES?
    new tree_type(BODIES,   SOFTENING==individual,pc,Ncrit,DMAX) :
    new tree_type(FL,X,M,Nb,SOFTENING==individual,pc,Ncrit,DMAX);
  STATE=built;                                     // set tree state            
}
//------------------------------------------------------------------------------
void falcON::reuse()
{
  // This routine uses the existing tree and                                    
  // -  re-sets the souls' positions, masses, and source flags                  
  // -  re-evaluates the basic source properties of the cells                   
  // and pre-compute basic source properties of cells                           
  if(TREE == 0) {
    warning(" cannot reuse(), no old tree existent");
    return grow();
  }
  TREE->reuse();
  STATE=built;                                     // set tree state            
}
//=============================================================================+
//                                                                             |
// 3 INFORMATION AND STATISTICS                                                |
// ============================                                                |
//                                                                             |
//=============================================================================+
const char* falcON::opening_method() const {
  return OPENING->method();
}
//------------------------------------------------------------------------------
real falcON::theta(const real m) const {
  return OPENING->theta(m);
}
//------------------------------------------------------------------------------
const nbdy::uint &falcON::No_bodies() const {
  return BODIES->N_body();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::No_bodies_used() const {
  return TREE->N_souls();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::No_cells_used() const {
  return TREE->N_cells();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::No_zombie_cells() const {
  return TREE->N_zombies();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::No_zombie_bodies() const {
  register uint  n=0;
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si) if(!is_in_tree(BODIES->flag(mybody(Si)))) n++;
  } else {
    LoopSouls(tree_type,TREE,Si) if(!is_in_tree(FL[mybody(Si)])) n++;
  }
  return n;
}
//------------------------------------------------------------------------------
namespace nbdy {
  inline void dump_nodes(const cell*Ci, ostream*oc, ostream*os,
			 const cell*C0, const soul*S0)
    // RECURSIVE
  {
    if(oc) { Ci->dump(*oc,S0,C0); *oc<<"\n"; }
    LoopCellKids(cell,Ci,c) dump_nodes(c,oc,os,C0,S0);
    if(os) LoopSoulKids(cell,Ci,s) { s->dump(*os,S0); *os<<"\n"; }
  }
  //----------------------------------------------------------------------------
  inline int trick(const real x, const int w)
  {
    if(x<0.001)  return util::max(1,w-5);
    if(x<0.01)   return util::max(1,w-4);
    if(x<0.1)    return util::max(1,w-3);
    if(x<one)    return util::max(1,w-2);
    if(x<ten)    return util::max(1,w-1);
    return w;
  }
}
//------------------------------------------------------------------------------
void falcON::dump_nodes(const char* fcells, const char* fleafs) const
{
  ofstream *oc = fcells? new ofstream(fcells) : 0;
  ofstream *os = fleafs? new ofstream(fleafs) : 0;
  if(oc) { cell::dump_head(*oc); (*oc)<<"\n"; }
  if(os) { soul::dump_head(*os); (*os)<<"\n"; }
  if(oc || os) ::dump_nodes(root(),oc,os,
			    TREE->begin_cells(), TREE->begin_souls());
  if(oc) { oc->flush(); delete oc; }
  if(os) { os->flush(); delete os; }
}
//------------------------------------------------------------------------------
const char* falcON::describe(const kern_type KER)
{
  switch(KER) {
  case p0: return "P0==Plummer";
  case p1: return "P1";
  case p2: return "P2";
  case p3: return "P3";
  case newton:
  default: return "Newtonian";
  }
}
//------------------------------------------------------------------------------
void falcON::stats(ostream& out) const
{
  out<<"\n state:                ";
  if(STATE&unset)  out<<" unset";
  if(STATE&built)  out<<" built";
  if(STATE&reused) out<<" reused";
  out<<"\n";
  if(STATE!=unset) {
    out<<" root center:           "<<center(root())       <<"\n"
       <<" root radius:           "<<radius(root())       <<"\n"
       <<" bodies loaded:         "<<number(root())       <<"\n"
       <<" total mass:            "<<mass  (root())       <<"\n"
       <<" N_crit:                "<<Ncrit                <<"\n"
       <<" cells used:            "<<TREE->N_cells()      <<"\n"
       <<" zombies elimitated:    "<<TREE->N_zombies()    <<"\n"
       <<" maximum depth:         "<<TREE->depth()        <<"\n"
       <<" current theta:         "<<OPENING->theta_min() <<"\n"
       <<" current opening:       "<<OPENING->method()    <<"\n";
    if(SOFTENING == global) 
      out<<" softening:             global\n"
	 <<" softening length:      "<< EPS               <<"\n";
    else
      out<<" softening:             individual\n";
    out<<" softening kernel:      "<<describe_kernel()    <<"\n"
       <<" Taylor coeffs used:    "<<Ncoeffs              <<"\n";
  }
  register uint total = STATS->BB_iacts()
                      + STATS->CB_taylor_iacts()
                      + STATS->CB_direct_iacts()
                      + STATS->CC_taylor_iacts()
                      + STATS->CC_direct_iacts()
                      + STATS->CS_direct_iacts(), part;
  register real percent;
  out<<" interaction statitics:\n"
     <<"     type          approx   direct      total\n"
     <<" # body-body :          - ";
  part    = STATS->BB_iacts();
  percent = 100.*part/real(total);
  out<<setw( 8)<<STATS->BB_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-body : ";
  part    = STATS->CB_taylor_iacts()
	   +STATS->CB_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw(10)<<STATS->CB_taylor_iacts()<<" "
     <<setw( 8)<<STATS->CB_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-cell : ";
  part    = STATS->CC_taylor_iacts()
	   +STATS->CC_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw(10)<<STATS->CC_taylor_iacts()<<" "
     <<setw( 8)<<STATS->CC_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # cell-self :          - ";
  part    = STATS->CS_direct_iacts();
  percent = 100.*part/real(total);
  out<<setw( 8)<<STATS->CS_direct_iacts()<<" "
     <<setw(10)<<part<<" = "
     <<setprecision(trick(percent,5))<<setw(8)<<percent<<"%\n"
     <<" # total     : ";
  part=STATS->BB_iacts()
      +STATS->CB_direct_iacts()
      +STATS->CC_direct_iacts()
      +STATS->CS_direct_iacts();
  out<<setw(10)<<STATS->CB_taylor_iacts()+STATS->CC_taylor_iacts()<<" "
     <<setw( 8)<<part<<" "
     <<setw(10)<<total<<" =  100.000%\n";
}
//------------------------------------------------------------------------------
nbdy::uint falcON::BB_interactions() const {
  return STATS->BB_iacts();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::MB_interactions() const {
  return STATS->CB_direct_iacts() +STATS->CC_direct_iacts()
        +STATS->CS_direct_iacts();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::CB_interactions() const {
  return STATS->CB_taylor_iacts();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::CC_interactions() const {
  return STATS->CC_taylor_iacts();
}
//------------------------------------------------------------------------------
nbdy::uint falcON::total_interactions () const {
  return BB_interactions()+MB_interactions()
        +CB_interactions()+CC_interactions();
}
//------------------------------------------------------------------------------
const nbdy::vect& falcON::root_center() const { return center(root()); }
const nbdy::real& falcON::root_radius() const { return radius(root()); }
const nbdy::uint& falcON::root_depth () const { return TREE->depth(); }
const nbdy::uint& falcON::root_number() const { return number(root()); }
const nbdy::real& falcON::root_mass  () const { return mass  (root()); }
//-----------------------------------------------------------------------------+
//                                                                             |
// 4 ESTIMATION OF DENSITY AND SUCH                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace nbdy {
  inline void pass_flags_up(tree_type*T, const bodies*B)
  {
    LoopSouls  (tree_type,T,Si) Si->copy_sink_flag(B);
    LoopCellsUp(tree_type,T,Ci) Ci->update_sink_flag();
  }
  //----------------------------------------------------------------------------
  inline void pass_flags_up(tree_type*T, const int *FL)
  {
    LoopSouls  (tree_type,T,Si) Si->copy_sink_flag(FL);
    LoopCellsUp(tree_type,T,Ci) Ci->update_sink_flag();
  }
  //----------------------------------------------------------------------------
  inline void guess_N(cell*Ci, real n, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci)/(NSUB*util::cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci)/(NSUB*util::square(radius(Ci)));
#endif // NDIM==3
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_N(c,n,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = n;
  }
  //----------------------------------------------------------------------------
  inline void guess_sd(cell*Ci, real sd, const uint& Nx)
  {
    if(number(Ci)>Nx || sd==zero) sd = mass(Ci) / (4*util::square(radius(Ci)));
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_sd(c,sd,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = sd;
  }
  //----------------------------------------------------------------------------
  inline void guess_rho(cell*Ci, real rh, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci)/(NSUB*util::cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci)/(NSUB*util::square(radius(Ci)));
#endif // NDIM==3
    LoopCellKids(cell,Ci,c) if(is_sink(c)) guess_rho(c,rh,Nx);
    LoopSoulKids(cell,Ci,s) if(is_sink(s)) s->rho() = rh;
  }
}
//------------------------------------------------------------------------------
void falcON::estimate_n(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_N(root(),zero,Nx);                         // guess n_i                 
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(BODIES);     // copy n_i into body        
  } else {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
  }
}
//------------------------------------------------------------------------------
void falcON::estimate_sd(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_sd(root(),zero,Nx);                        // guess SD_i                
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(BODIES);     // copy n_i into body        
  } else {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
  }
}
//------------------------------------------------------------------------------
void falcON::estimate_rho(const uint Nx)
{
  if(BODIES) pass_flags_up(TREE,BODIES);
  else       pass_flags_up(TREE,FL);               // update flags              
  guess_rho(root(),zero,Nx);                       // guess n_i                 
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(BODIES);     // copy n_i into body        
  } else {
    LoopSouls(tree_type,TREE,Si)
      if(is_sink(Si)) Si->update_dens(RH);         // copy n_i into array RH    
  }
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 5 APPROXIMATION OF GRAVITY                                                  |
// ==========================                                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace nbdy {
  //----------------------------------------------------------------------------
  // preparation: passing flags & eps up the tree, set R_crit                   
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T)
  {
    register uint n=0u;
    if(SOFTENING == individual) {
      register real Emin, Emax;
#ifdef SOFT_MEAN
      register real Eps;
#endif
      LoopCellsUp(tree_type,T,Ci) {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->reset_sink_flag();
	Emin = zero;
	Emax = zero;
#ifdef SOFT_MEAN
	Eps  = zero;
#endif
	LoopCellKids(cell,Ci,c) {
	  Ci->add_sink_flag(c);
	  if(Emin==zero || emin(c)<Emin) Emin = emin(c);
	  if(Emax==zero || emax(c)>Emax) Emax = emax(c);
#ifdef SOFT_MEAN
	  Eps += eps(c);
#endif
	}
	LoopSoulKids(cell,Ci,s) {
	  Ci->add_sink_flag(s);
	  if(Emin==zero || eps(s)<Emin) Emin = eps(s);
	  if(Emax==zero || eps(s)>Emax) Emax = eps(s);
#ifdef SOFT_MEAN
	  Eps += mass(s)*eps(s);
#endif
	}
	if(is_sink(Ci)) n++;
	Ci->emin() = Emin;
	Ci->emax() = Emax;
#ifdef SOFT_MEAN
	Ci->eps()  = Eps;
#else
	Ci->eps()  = Emin;
#endif
      }
#ifdef SOFT_MEAN
      LoopCellsUp(tree_type,T,Ci) Ci->eps() *= imass(Ci);
#endif
    } else { // SOFTENING == fixed, no cell-intern epsila
      LoopCellsUp(tree_type,T,Ci) {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->update_sink_flag();
	if(is_sink(Ci)) n++;
      }
    }
    return n;
  }
  //----------------------------------------------------------------------------
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T,
				    const bodies    *B) {
    LoopSouls(tree_type,T,Si) Si->reset_grav(B);
    return pass_flags_and_eps_up(OPENING,SOFTENING,T);
  }
  //----------------------------------------------------------------------------
  inline uint pass_flags_and_eps_up(const opening   *OPENING,
				    const soft_type &SOFTENING,
				          tree_type *T,
				    const int       *FL,
				    const areal     *EP=0) {
    if(EP) LoopSouls(tree_type,T,Si) Si->reset_grav(FL,EP);
    else   LoopSouls(tree_type,T,Si) Si->reset_grav(FL);
    return pass_flags_and_eps_up(OPENING,SOFTENING,T);
  }
  //----------------------------------------------------------------------------
}
//------------------------------------------------------------------------------
// approximation: prepare, interact, and evaluate gravity                       
void falcON::approximate_gravity(bool split, const uint ndirect[4]
#ifdef __GRAPE
				 , bool grape
#endif
				 )
{
#ifdef TEST_CPU
  register clock_t  cpu0 = clock(), cpu1;
#endif
  // 1. preparation phase                                                       
  register uint NS = BODIES?                       // IF(using bodies)         >
    pass_flags_and_eps_up(OPENING,SOFTENING,       //   copy flags & eps_i from 
			  TREE,BODIES) :           //   bodies and pass up tree 
    pass_flags_and_eps_up(OPENING,SOFTENING,       // ELSE copy them from arrays
			  TREE,FL,EP);             //   and pass up tree        
  if(NS==0) {                                      // IF(no sink cells)        >
    warning(" falcON::approximate_gravity(): no body wants updating");
    return;                                        //   we are done already     
  }                                                // <                         
  TREE->set_poles(3);
#ifdef TEST_CPU
  cpu1 = clock();
  std::cerr<<
    " preparation took "<<(cpu1 - cpu0) / real(CLOCKS_PER_SEC)<<" sec \n";
  cpu0 = cpu1;
#endif
  // 2. interaction and evaluation phases                                       
  STATS->reset();                                  // reset iaction statistics  
  register uint NP = split? 4+NS/8 : NS;           // initial size of C_i pool  
  gravity_kernel GK(STATS,KERNEL,NP,EPS,SOFTENING, // init gravity kernel       
		    ndirect,false,3
#ifdef __GRAPE
		    ,grape? TREE->N_souls() : 0
		    ,grape? TREE->begin_souls() : 0
#endif
		    );
  // case 2.1: interweave interaction and evaluation phase to save some memory  
#ifdef TEST_CPU
  cpu1 = clock();
  register double ti=0.,te=(cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cpu0 = cpu1;
#endif
#ifdef __GRAPE
  if(split && (!grape || is_bough(root())) ) {
#else
  if(split) {                                      // IF(interweaving)         >
#endif
    MutualInteractor<gravity_kernel>
      MI(&GK,TREE->depth()-1);                     //   init mutual interactor  
    LoopCellKids(cell,root(),c1) {                 //   loop cell kids c1      >
      LoopCellSecd(cell,root(),c1,c2)              //     loop cell kids c2>=c1 
	MI.cell_cell(c1,c2);                       //       interaction phase   
      LoopSoulKids(cell,root(),s2)                 //     loop soul kids        
	MI.cell_soul(c1,s2);                       //       interaction phase   
#ifdef TEST_CPU
      cpu1 = clock();
      ti  += (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
      cpu0 = cpu1;
#endif
      GK.evaluate_grav(c1);                        //     evaluation phase      
#ifdef TEST_CPU
      cpu1 = clock();
      te  += (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
      cpu0 = cpu1;
#endif
    }                                              //   <                       
    LoopSoulKids(cell,root(),s1) {                 //   loop soul kids s1      >
      LoopSoulSecd(cell,root(),s1+1,s2)            //     loop soul kids s2 > s1
	GK.interact(s1,s2);                        //       interaction phase   
      s1->normalize_grav();                        //     evaluation phase      
    }                                              //   <                       
#ifdef TEST_CPU
    cpu1 = clock();
    ti  += (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
    cpu0 = cpu1;
#endif
  // case 2.2: non-interweaving, as in the JCP paper                            
  } else {                                         // < ELSE(not splitting)    >
#ifdef __GRAPE
    if(split) warning(" cannot interweave interaction & evaluation");
#endif
    MutualInteractor<gravity_kernel>
      MI(&GK,TREE->depth());                       //   init mutual interactor  
    MI.cell_cell(root(),root());                   //     interaction phase     
#ifdef TEST_CPU
    cpu1 = clock();
    ti  += (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
    cpu0 = cpu1;
#endif
    GK.evaluate_grav(root());                      //     evaluation phase      
#ifdef TEST_CPU
    cpu1 = clock();
    te  += (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
    cpu0 = cpu1;
#endif
  }                                                // <                         
  Ncoeffs = GK.coeffs_used();                      // remember # coeffs used    
#ifdef TEST_CPU
  std::cerr<<
    " interaction took "<<ti<<" sec\n"
    " evaluation  took "<<te<<" sec\n";
#endif
  // 3. copy gravity from the souls into the bodies                             
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) Si->update_grav(BODIES);
  } else {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) Si->update_grav(A,PH);
  }
#ifdef TEST_CPU
  cpu1 = clock();
  std::cerr<<
    " copying     took "<<(cpu1 - cpu0) / real(CLOCKS_PER_SEC)<<" sec\n";
#endif
}
//------------------------------------------------------------------------------
// exact gravity by direct summation                                            
void falcON::exact_gravity()
{
  register uint Nu = pass_flags_and_eps_up(OPENING,SOFTENING,TREE,FL,EP);
  if(Nu==0) {
    warning(" tree::exact_gravity(): no body wants updating");
    return;
  }
  STATS->reset();
  gravity_kernel K(STATS,KERNEL,0,EPS,SOFTENING);
  K.direct_summation(root());
  if(BODIES) {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) {
      Si->normalize_grav();
      Si->update_grav(BODIES);
    }
  } else {
    LoopSouls(tree_type,TREE,Si) if(is_sink(Si)) {
      Si->normalize_grav();
      Si->update_grav(A,PH);
    }
  }
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 6 NEIGHBOUR FINDING AND COUNTING                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
void falcON::make_iaction_list(      elem_pair*bl,
			       const uint      nl,
			             uint     &na,
			       const real      tau)
{
  if(BODIES == 0) error(" mismatch in using bodies/arrays");
  if(tau < zero) {             // SPH
    LoopSouls(tree_type,TREE,Si) if(is_sph(BODIES->flag(mybody(Si)))) {
      flag_for_subtree(Si);
      Si->add_body_flag(BODIES);
    }
    sticky_tree      stree(TREE);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    MutualInteractor<neighbour_finder> MI(&sfind,stree.depth());
    MI.cell_self(stree.root());
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls(tree_type,TREE,Si) if(is_sticky(BODIES->flag(mybody(Si)))) {
      flag_for_subtree(Si);
      Si->add_body_flag(BODIES);
    }
    sticky_tree   stree(TREE);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    MutualInteractor<sticky_finder> MI(&sfind,stree.depth());
    MI.cell_self(stree.root());
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void falcON::count_sph_neighbours()
{
  if(BODIES == 0) error(" mismatch in using bodies/arrays");
  LoopSouls(tree_type,TREE,Si) if(is_sph(BODIES->flag(mybody(Si)))) {
    flag_for_subtree(Si);
    Si->add_body_flag(BODIES);
  }
  sticky_tree       stree (TREE);
  neighbour_counter scount;
  stree.prepare_sph();
  MutualInteractor<neighbour_counter> MI(&scount,stree.depth());
  MI.cell_self(stree.root());
  stree.set_num(TREE->begin_souls());
}
//-----------------------------------------------------------------------------+
void falcON::make_iaction_list(      elem_pair*bl,
			       const uint      nl,
			             uint     &na,
			       const areal    *s,
			       const areal    *v[NDIM],
			       const real      tau)
{
  if(BODIES) error(" mismatch in using bodies/arrays");
  if(tau < zero) {             // SPH
    LoopSouls(tree_type,TREE,Si) if(is_sph(FL[mybody(Si)])) {
      flag_for_subtree(Si);
      Si->add_body_flag(FL);
    }
    sticky_tree      stree(TREE,s,v);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    MutualInteractor<neighbour_finder> MI(&sfind,stree.depth());
    MI.cell_self(stree.root());
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls(tree_type,TREE,Si) if(is_sticky(FL[mybody(Si)])) {
      flag_for_subtree(Si);
      Si->add_body_flag(FL);
    }
    sticky_tree   stree(TREE,s,v);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    MutualInteractor<sticky_finder> MI(&sfind,stree.depth());
    MI.cell_self(stree.root());
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void falcON::count_sph_neighbours(const areal* s, areal* NUM)
{
  if(BODIES) error(" mismatch in using bodies/arrays");
  LoopSouls(tree_type,TREE,Si) if(is_sph(FL[mybody(Si)])) {
    flag_for_subtree(Si);
    Si->add_body_flag(FL);
  }
  sticky_tree       stree(TREE,s,0);
  neighbour_counter scount;
  stree.prepare_sph();
  MutualInteractor<neighbour_counter> MI(&scount,stree.depth());
  MI.cell_self(stree.root());
  stree.set_num(NUM);
}
////////////////////////////////////////////////////////////////////////////////
