//-----------------------------------------------------------------------------+
//                                                                             |
// grat.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// implementing nbdy/grat.h                                                    |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <nbdy/grat.h>
#include <nbdy/tayl.h>
#include <util/inln.h>
using namespace util;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // auxiliary stuff                                                            
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  inline real bmax(const cell*C)
    // This routines returns the distance from the cell's cofm to its most      
    // distant corner.                                                          
  {
    register real bq=zero;                         // initialize return variable
    for(register indx i=0; i<NDIM; i++)            // loop dimensions           
      bq += square(radius(C)+abs(cofm(C)[i]-center(C)[i])); // add in square    
    return SQRT(bq);                               // return sqrt of total      
  }
  //----------------------------------------------------------------------------
  template<bool REUSE>
  inline void eval_basic_source(const gravity_tree *T)
    // This routines computes for all cells of tree T:                          
    //  - mass, 1/mass, cofm, rmax (needed for almost everything)               
    // To be uses AFTER souls have been updated                                 
  {
    register real  mon;                            // monopole                  
    register vect  dip;                            // dipole                    
    LoopCellsUp(gravity_tree,T,Ci) {               // loop cells upwards >      
      mon = zero;                                  //   reset monopole          
      dip = zero;                                  //   reset dipole            
      LoopCellKids(cell,Ci,c) {                    //   loop sub-cells c >      
	mon += mass(c);                            //       sum up monopole     
	dip += cofm(c);                            //       sum up dipole       
      }                                            //   <                       
      LoopSoulKids(cell,Ci,s) {                    //   loop sub-souls s >      
	mon += mass(s);                            //       sum up monopole     
	dip += mass(s)*cofm(s);                    //       sum up dipole       
      }                                            //   <                       
      Ci->mass() = mon;                            //   set mass                
      Ci->imass()= (mon==zero)? zero:one/mon;      //   set 1/mass              
      Ci->cofm() = dip;                            //   set dipole              
    }                                              // <                         
    register real dmax, x, Xq;                     // d_max, auxiliary vars     
    LoopCellsUp(gravity_tree,T,Ci) {               // loop cells upwards >      
      Ci->cofm() *= imass(Ci);                     //   cofm = dipole/mass      
      dmax = zero;                                 //   reset dmax              
      LoopSoulKids(cell,Ci,s)                      //   loop sub-souls s >      
	update_max(dmax,dist_sq(cofm(s),cofm(Ci)));//     update d_max^2 <      
      if(has_soul_children(Ci)) dmax = SQRT(dmax); //   d_max due to sub-souls  
      LoopCellKids(cell,Ci,c) {                    //   loop sub-cells c >      
	Xq = dist_sq(cofm(c),cofm(Ci));            //     distance^2            
	x  = dmax - rmax(c);                       //     d=distance + r_i      
	if(zero > x || Xq>x*x)                     //     IF(d>d_max)      >    
	  dmax = SQRT(Xq) + rmax(c);               //       set d_max = d  <    
      }                                            //   <                       
      Ci->rmax() = REUSE? dmax:min(dmax,bmax(Ci)); //   assign r_max            
    }                                              // <                         
  }
  //----------------------------------------------------------------------------
  inline int Npoles(const unsigned order) {
    return (order > 2)?                            // return number of scalars  
      TaylorExpansion::Ncoeff(order-1) -           // needed in the multipoles  
      TaylorExpansion::Ncoeff(1)                   // at given order            
      : 0;
  }
  //----------------------------------------------------------------------------
  inline void evaluate_poles(gravity_tree*T, const int P) {
    // computes proper multipole moments up to order P-1 for all cells in tree  
    register vect Xi;                              // distance vector           
    if       (P==3) {                              // set M2 only               
      SYM2(M2);                                    //   macro in tens.h         
      LoopCellsUp(gravity_tree,T,Ci) {             //   loop tree upwards >     
	M2 = zero;                                 //     reset M2 = 0          
	LoopSoulKids(cell,Ci,s) {                  //     loop over soul kids > 
	  Xi  = cofm(s)-cofm(Ci);                  //       Xi = Yi - Z         
	  M2.add_outer_prod(Xi,mass(s));           //       M2 += Xi2 Mi2       
	}                                          //     <                     
	LoopCellKids(cell,Ci,c) {                  //     loop over cell kids > 
	  M2 += quad(c);                           //       M2 += Mi2           
	  Xi  = cofm(c)-cofm(Ci);                  //       Xi = Zi - Z         
	  M2.add_outer_prod(Xi,mass(c));           //       M2 += Xi2 Mi2       
	}                                          //     <                     
	Ci->quad() = M2;                           //     assign quadrupole     
      }                                            //   <                       
    } else if(P==4) {                              // set M2, M3                
      SYM2(M2); SYM2(X2);                          //   macro in tens.h         
      SYM3(M3);                                    //   macro in tens.h         
      LoopCellsUp(gravity_tree,T,Ci) {             //   loop tree upwards >     
	M2 = zero;                                 //     reset M2 = 0          
	M3 = zero;                                 //     reset M3 = 0          
	LoopSoulKids(cell,Ci,s) {                  //     loop over soul kids > 
	  Xi = cofm(s)-cofm(Ci);                   //       Xi  = Yi - Z        
	  M2+= X2.outer_prod(Xi,mass(s));          //       M2 += Xi2 Mi        
	  M3.add_outer_prod(X2,Xi);                //       M3 += Xi3 Mi        
	}                                          //     <                     
	LoopCellKids(cell,Ci,c) {                  //     loop over cell kids > 
	  M2+= quad(c);                            //       M2 += Mi2           
	  M3+= octo(c);                            //       M3 += Mi3           
	  Xi = cofm(c)-cofm(Ci);                   //       Xi  = Zi - Z        
	  M2+= X2.outer_prod(Xi,mass(c));          //       M2 += Xi2 Mi        
	  M3.add_outer_prod(X2,Xi);                //       M3 += Xi3 Mi        
	  M3.add_symm_prod(quad(c),Xi);            //       M3 += Xi % Mi2      
	}                                          //     <                     
	Ci->quad() = M2;                           //     assign quadrupole     
	Ci->octo() = M3;                           //     assign octopole       
      }                                            //   <                       
    } else if(P==5) {                              // set M2, M3, M4            
      SYM2(M2); SYM2(X2);                          //   macro in tens.h         
      SYM3(M3); SYM3(X3);                          //   macro in tens.h         
      SYM4(M4);                                    //   macro in tens.h         
      LoopCellsUp(gravity_tree,T,Ci) {             //   loop tree upwards >     
	M2 = zero;                                 //     reset quadrupole      
	M3 = zero;                                 //     reset octopole        
	M4 = zero;                                 //     reset hexadecupole    
	LoopSoulKids(cell,Ci,s) {                  //     loop over soul kids > 
	  Xi = cofm(s)-cofm(Ci);                   //       Xi  = Yi - Z        
	  M2+= X2.outer_prod(Xi,mass(s));          //       M2 += Xi2 Mi        
	  M3+= X3.outer_prod(X2,Xi);               //       M3 += Xi3 Mi        
	  M4.add_outer_prod(X3,Xi);                //       M4 += Xi4 Mi        
	}                                          //     <                     
	LoopCellKids(cell,Ci,c) {                  //     loop over cell kids > 
	  M2+= quad(c);                            //       M2 += Mi2           
	  M3+= octo(c);                            //       M3 += Mi3           
	  M4+= hexa(c);                            //       M4 += Mi4           
	  Xi = cofm(c)-cofm(Ci);                   //       Xi  = Zi - Z        
	  M2+= X2.outer_prod(Xi,mass(c));          //       M2 += Xi2 Mi        
	  M3+= X3.outer_prod(X2,Xi);               //       M3 += Xi3 Mi        
	  M4.add_outer_prod(X3,Xi);                //       M4 += Xi4 Mi        
	  M3+= X3.outer_prod(quad(c),three*Xi);    //       M3 += 3 Xi Mi2      
	  M4.add_outer_prod(X3,two*Xi);            //       M4 += 6 Xi2 Mi2     
	  M4.add_outer_prod(octo(c),four*Xi);      //       M4 += 4 Xi Mi3      
	}                                          //     <                     
	Ci->quad() = M2;                           //     assign quadrupole     
	Ci->octo() = M3;                           //     assign octopole       
	Ci->hexa() = M4;                           //     assign hexadecupole   
      }                                            //   <                       
    } else if(P >5) {
      error(" order P > 5 not supported");
    }
  }
  //----------------------------------------------------------------------------
  inline void normalize_poles(gravity_tree*T, const int P) {
    // transforms multipoles into reduced & normlized ~M^n = M^n / (n! * M)     
    if       (P==3) {
      LoopCellsDown(gravity_tree,T,Ci)             // loop cells                
	Ci->quad() *= if2*imass(Ci);               //   normalize quadrupole <  
    } else if(P==4) {
      LoopCellsDown(gravity_tree,T,Ci) {           // loop cells                
        Ci->quad() *= if2*imass(Ci);               //   normalize quadrupole    
	Ci->octo() *= if3*imass(Ci);               //   normalize octopole      
      }
    } else if(P==5) {
      LoopCellsDown(gravity_tree,T,Ci) {           // loop cells                
        Ci->quad() *= if2*imass(Ci);               //   normalize quadrupole    
	Ci->octo() *= if3*imass(Ci);               //   normalize octopole      
	Ci->hexa() *= if4*imass(Ci);               //   normalize hexadecupole  
      }
    } else if(P >5) {
      error(" order P > 5 not supported");
    }
  }    
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class nbdy::gravity_tree                                                   
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
#define LoopMyCellsUp(NAME)                        /* name for cell* s       */\
  for(register cell_type	                   /* type of cell           */\
     *NAME  = rbegin_cells();                      /* from last cell         */\
      NAME != rend_cells();                        /* until beyond root cell */\
    --NAME)                                        // get previous cell         
  //////////////////////////////////////////////////////////////////////////////
#define LoopMyCellsDown(NAME)                      /* name for cell* s       */\
  for(register cell_type	                   /* type of cell           */\
     *NAME  = begin_cells();                       /* from last cell         */\
      NAME != end_cells();                         /* until beyond root cell */\
    ++NAME)                                        // get previous cell         
  //////////////////////////////////////////////////////////////////////////////
#define LoopMySouls(NAME)                          /* name for cell* s       */\
  for(register soul_type	                   /* type of cell           */\
     *NAME  = begin_souls();                       /* from last cell         */\
      NAME != end_souls();                         /* until beyond root cell */\
    ++NAME)                                        // get previous cell         
  //----------------------------------------------------------------------------
  gravity_tree::gravity_tree(const bodies_type*b,  // I: bodies                 
			     const bool        e,  // [I: allow for cell's eps] 
			     const vect       *c,  // [I: root center]          
			     const uint        n,  // [I: N_crit]               
			     const uint        d)  // [I: max tree depth]       
    : tree<cell>(b,c,n,d), M(0), P(0),
    nP    ( -1 ), 
    nS    ( e? cell::N_stuff() : cell::N_eps() ),
    POLES ( 0 ),
    STUFF (new scalar_type[N_cells()*nS] )         // allocate memory for stuff 
  {
    register scalar_type*Si=STUFF;                 // pter to cell's stuff      
    LoopMyCellsDown(Ci) {                          // loops cells         >     
      Ci->STUFF = Si;                              //   give memory to cell     
      Si += nS;                                    //   pter to next memory     
    }                                              // <                         
    eval_basic_source<false>(this);
  }
  //----------------------------------------------------------------------------
  gravity_tree::gravity_tree(const flag_type*f,    // I: array with flags       
			     const areal  *x[NDIM],// I: arrays with x,y,z      
			     const areal  *m,      // I: array with masses      
			     const uint    nb,     // I: size of arrays         
			     const bool    e,      // [I: allow for cell's eps] 
			     const vect   *c,      // [I: root center]          
			     const uint    n,      // [I: N_crit]               
			     const uint    d)      // [I: max tree depth]       
    : tree<cell>(f,x,nb,c,n,d), M(m), P(0),
    nP    ( -1 ), 
    nS    ( e? cell::N_stuff() : cell::N_eps() ),
    POLES ( 0 ),
    STUFF (new scalar_type[N_cells()*nS] )         // allocate memory for stuff 
  {
    register scalar_type*Si=STUFF;                 // pter to cell's stuff      
    LoopMyCellsDown(Ci) {                          // loops cells         >     
      Ci->STUFF = Si;                              //   give memory to cell     
      Si += nS;                                    //   pter to next memory     
    }                                              // <                         
    LoopMySouls(Si) Si->set_mass(M); 
    eval_basic_source<false>(this);
  }
  //----------------------------------------------------------------------------
  void gravity_tree::reuse() const {
    if(use_bodies()) LoopMySouls(Si) Si->set_up(my_bodies());
    else             LoopMySouls(Si) Si->set_up(M,my_pos());
    eval_basic_source<true>(this);
  }
  //----------------------------------------------------------------------------
  void gravity_tree::set_poles(const unsigned p) {
    if(p!=P || POLES==0) {                         // IF(need memory for poles)>
      if(POLES) { delete[] POLES; POLES=0; }       //   free old memory         
      P  = p;                                      //   set order               
      nP = Npoles(P);                              //   scalars needed per cell 
      if(nP > 0) {                                 //   IF(any memory needed)  >
	POLES = new scalar_type[N_cells()*nP];     //     allocate memory       
	register scalar_type*Pi=POLES;             //     pter to cell's poles  
	LoopMyCellsDown(Ci) {                      //     loops cells         > 
	  Ci->POLES = Pi;                          //       give memory to cell 
	  Pi += nP;                                //       pter to next memory 
	}                                          //     <                     
      }                                            //   <                       
    }                                              // <                         
    evaluate_poles (this,P);                       // set Mi^(n), n=2,...P-1    
    normalize_poles(this,P);                       // normalize poles           
  }
  //----------------------------------------------------------------------------
#undef LoopMySouls
#undef LoopMyCellsUp
#undef LoopMyCellsDown
////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
