//-----------------------------------------------------------------------------+
//                                                                             |
// grav.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// implementing nbdy/grav.h                                                    |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <nbdy/grav.h>
#ifdef __GRAPE
#  include <nbdy/grap.h>
#  include <nbdy/graa.h>
#endif
//------------------------------------------------------------------------------
#include <nbdy/gren.h>
#include <util/inln.h>
using namespace util;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // methods eps_mutual()                                                       
  //                                                                            
  // return, given two nodes, the softening length to be used in their mutual   
  // interaction. The behaviour is steeredd by the macro SOFT_MIN, usually set  
  // by a compiler flag.                                                        
  // These methods are used ONLY when no global softening length is used.       
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
#ifdef SOFT_MIN
  // 1. eps_mutual = min (eps_i, eps_j)                                         
  inline real eps_mutual(soul*A, soul*B) { return min(eps(A),eps(B)); }
  inline real eps_mutual(cell*A, soul*B) { return min(emin(A),eps(B)); }
  inline real eps_mutual(cell*A, cell*B) { return min(emin(A),emin(B)); }
#else
  // 2. eps_mutual = mass-weighted mean of eps_i and eps_j                      
  inline real eps_mutual(soul*A, soul*B) {
    if(eps(A)  == eps(B))  return eps(A);
    if(mass(A) == mass(B)) return half*(eps(A)+eps(B));
    return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
  inline real eps_mutual(cell*A, soul*B) {
    if(eps(A)==eps(B)) return eps(A);
    else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
  inline real eps_mutual(cell*A, cell*B) {
    if(eps(A)==eps(B)) return eps(A);
    else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  //----------------------------------------------------------------------------
#endif
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // methods emin() & emax()                                                    
  //                                                                            
  // return the minimum of two nodes' individual softening lengths.             
  // These methods are used ONLY when no global softening length is used.       
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  inline real emin(const cell *a, const cell *b) { return min(emin(a),emin(b));}
  inline real emin(const cell *a, const soul *b) { return min(emin(a),eps(b)); }
  inline real emax(const cell *a, const cell *b) { return max(emax(a),emax(b));}
  inline real emax(const cell *a, const soul *b) { return max(emax(a),eps(b)); }
}
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// class nbdy::gravity_kernel                                                   
//                                                                              
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// construction & destruction                                                   
//------------------------------------------------------------------------------
gravity_kernel::
gravity_kernel(      TreeStats*t,                 // I: statistics              
	       const kern_type k,                 // I: type of kernel          
	       const uint      np,                // I: initial pool size       
	       const real      e,                 //[I: softening length]       
	       const soft_type s,                 //[I: type of softening]      
	       const uint      nd[4],             //[I: direct sum control]     
	       const bool      full_pot,          //[I: use Pth pole in pot]
	       const uint      p                  //[I: expansion order]        
#ifdef __GRAPE
	      ,const uint      ns                 //[I; GRAPE?: # souls]        
	      ,const soul     *s0                 //[I: GRAPE?: souls]          
#endif
		     ) :
  TaylorExpansion ( np, p ),                      // TaylorSeries stuff         
  FULL_POT        ( full_pot ),                   // Pth pole in pot?           
#ifdef __GRAPE
  GARR ( s==global && ns ?                        // IF global eps AND wanted   
	 new grape_arrays(ns,s0) : 0 ),           // use grape for direct       
#endif
  SOFT ( s ),                                     // set softening type         
  KERN ( k ),                                     // set softening kernel       
  STAT ( t ),                                     // set statistics             
  EPS  ( e ),                                     // set softening length       
  EPQ  ( EPS*EPS ),                               // set eps^2                  
  IEP  ( (EPS>0)? one/EPS : zero ),               // set 1/eps                  
  IEQ  ( IEP*IEP )                                // set 1/eps^2                
{
  N_PRE[0]  = nd[0];                              // C-B direct before try      
  N_PRE[1]  = 0u;                                 // C-C direct before try      
  N_PRE[2]  = nd[3];                              // C-S direct before try      
  N_POST[0] = nd[1];                              // C-B direct after fail      
  N_POST[1] = nd[2];                              // C-C direct after fail      
  N_POST[2] = 0u;                                 // C-S direct after fail      
#ifdef __GRAPE
  if(KERN != p0)
    cerr<<
      "\n falcON WARNING:  "
      "using different kernels for direct summation (with GRAPE)"
      "\n                  and approximate interactions\n"<<endl;
#endif
}
//------------------------------------------------------------------------------
gravity_kernel::~gravity_kernel() {               // destructor                 
#ifdef __GRAPE
  if(GARR) delete GARR;                           //   delete GRAPE arrays      
#endif
}
//------------------------------------------------------------------------------
// body-many bodies interaction for global and individual softening             
//------------------------------------------------------------------------------
#define LOOP_B for(register soul*B=B0; B!=BN; B++)
#define LOOPMANY							\
if(A_is_sink)								\
  switch(B_sink) {							\
  case all:  LOOP_B { SET_RAD MAMB SET_AB } break;			\
  case some: LOOP_B { if(is_sink(B)) { SET_RAD MAMB SET_AB }		\
                      else           { SET_RAD MAMB SET_A  } } break;	\
  case none: LOOP_B { SET_RAD MAMB SET_A } break;			\
  }									\
else									\
  switch(B_sink) {							\
  case all:  LOOP_B { SET_RAD MAMB SET_B } break;			\
  case some: LOOP_B if(is_sink(B)) { SET_RAD MAMB SET_B } break;	\
  case none: break;							\
  }
//------------------------------------------------------------------------------
real gravity_kernel::greens_function(vect&R) const
{
  register real Rq=norm(R),x,D0,ME;
  BODYBODY_GLOB(1);
  return D0;
}
//------------------------------------------------------------------------------
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with global softening length.          
//------------------------------------------------------------------------------
#define SET_RAD   R  = cofm(A)-cofm(B); Rq = norm(R);
#define MAMB      MM = mass(A)*mass(B);
#define GRAV_A    A->pot()-= D0; A->acc()+= R;
#define GRAV_B    B->pot()-= D0; B->acc()-= R;
#define SET_AB    { BODYBODY_GLOB(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_GLOB(MM) GRAV_A }
#define SET_B     { BODYBODY_GLOB(MM) GRAV_B }
//------------------------------------------------------------------------------
inline
void gravity_kernel::many_global(const bool A_is_sink, soul*A,
				 const who B_sink, soul*B0, const soul*BN) const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN, using a fixed softening length.                      
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
//------------------------------------------------------------------------------
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with individual softening lengths.     
//------------------------------------------------------------------------------
#undef SET_AB
#undef SET_A
#undef SET_B
#define SET_AB    { BODYBODY_INDI(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_INDI(MM) GRAV_A }
#define SET_B     { BODYBODY_INDI(MM) GRAV_B }
//------------------------------------------------------------------------------
inline
void gravity_kernel::many_individual(const bool A_is_sink, soul*A,
				     const who B_sink, soul*B0, const soul*BN)
  const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN, using individual softening lengths.                  
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
#undef  LOOPMANY
//------------------------------------------------------------------------------
inline gravity_kernel::who gravity_kernel::who_is_sink(const cell* C) const
  // This routines figures out whether all, none, or some of the bodies in cell 
  // C are sinks.                                                               
{
  if(!is_sink(C)) return none;
  if( al_sink(C)) return all;
                  return some;
}
//------------------------------------------------------------------------------
// replacing single cell-soul interaction with body-many body interactions      
//------------------------------------------------------------------------------
void gravity_kernel::many_global(cell* A, soul* B) const
{
    many_global(is_sink(B),B,who_is_sink(A),
		A->begin_soul_descendants(),A->end_soul_descendants());
}
//------------------------------------------------------------------------------
void gravity_kernel::many_individual(cell* A, soul* B) const
{
  many_individual(is_sink(B),B,who_is_sink(A),
		  A->begin_soul_descendants(),A->end_soul_descendants());
}
//------------------------------------------------------------------------------
// replacing single cell-cell interaction with body-many body interactions      
//------------------------------------------------------------------------------
#define LoopSoulDesc(C,S) for(register cell::soul_descendant		\
			      S =C->begin_soul_descendants();		\
			      S!=C->end_soul_descendants(); ++S)
//------------------------------------------------------------------------------
void gravity_kernel::many_global(cell* A, cell* B) const
{ 
#ifdef __GRAPE
  if(GARR) {
    register int 
      iA=GARR->index(A->begin_soul_descendants()),
      iB=GARR->index(B->begin_soul_descendants());
    mutual_grape_grav(EPS,
		      GRAPE_DATA_TYPE(util::min(xmin(A),xmin(B))),
		      GRAPE_DATA_TYPE(util::max(xmax(A),xmax(B))),
		      GRAPE_DATA_TYPE(mass(A->begin_soul_children())),
		      A->number(),
		      GARR->pos_arr(iA), GARR->mass_arr(iA),
		      al_sink(A)? 0 : GARR->sink_arr(iA),
		      GARR->acc_arr(iA), GARR->pot_arr(iA),
		      B->number(),
		      GARR->pos_arr(iB), GARR->mass_arr(iB),
		      al_sink(B)? 0 : GARR->sink_arr(iB),
		      GARR->acc_arr(iB), GARR->pot_arr(iB));
  } else {
#endif
    if(al_sink(A))
      LoopSoulDesc(A,Ai)
	many_global(true,Ai,who_is_sink(B),
		    B->begin_soul_descendants(),B->end_soul_descendants());
    else if(al_sink(B))
      LoopSoulDesc(B,Bi)
	many_global(true,Bi,who_is_sink(A),
		    A->begin_soul_descendants(),A->end_soul_descendants());
    else if(is_sink(A))
      LoopSoulDesc(A,Ai)
	many_global(is_sink(Ai),Ai,who_is_sink(B),
		    B->begin_soul_descendants(),B->end_soul_descendants());
    else if(is_sink(B))
      LoopSoulDesc(B,Bi)
	many_global(is_sink(Bi),Bi,who_is_sink(A),
		    A->begin_soul_descendants(),A->end_soul_descendants());
#ifdef __GRAPE
  }
#endif
}
//------------------------------------------------------------------------------
void gravity_kernel::many_individual(cell* A, cell* B) const
{
  if(al_sink(A))
    LoopSoulDesc(A,Ai)
      many_individual(true,Ai,who_is_sink(B),
		      B->begin_soul_descendants(),B->end_soul_descendants());
  else if(al_sink(B))
    LoopSoulDesc(B,Bi)
      many_individual(true,Bi,who_is_sink(A),
		      A->begin_soul_descendants(),A->end_soul_descendants());
  else if(is_sink(A))
    LoopSoulDesc(A,Ai)
      many_individual(is_sink(Ai),Ai,who_is_sink(B),
		      B->begin_soul_descendants(),B->end_soul_descendants());
  else if(is_sink(B))
    LoopSoulDesc(B,Bi)
      many_individual(is_sink(Bi),Bi,who_is_sink(A),
		      A->begin_soul_descendants(),A->end_soul_descendants());
}
//------------------------------------------------------------------------------
// replacing single cell self-interaction with body-many body interactions      
//------------------------------------------------------------------------------
#undef  LoopSoulDesc
#define LoopSoulDesc(C,S) for(register cell::soul_descendant		\
			      S =C->begin_soul_descendants();		\
			      S!=C->last_soul_descendant (); ++S)
//------------------------------------------------------------------------------
void gravity_kernel::many_global(cell* A) const
{
#ifdef __GRAPE
  if(GARR) {
    register int iA=GARR->index(A->begin_soul_descendants());
    self_grape_grav(EPS,
		    GRAPE_DATA_TYPE(xmin(A)),
		    GRAPE_DATA_TYPE(xmax(A)),
		    GRAPE_DATA_TYPE(mass(A->begin_soul_children())),
		    A->number(),
		    GARR->pos_arr(iA), GARR->mass_arr(iA),
		    al_sink(A)? 0 : GARR->sink_arr(iA),
		    GARR->acc_arr(iA), GARR->pot_arr(iA));
  } else {
#endif
    if(al_sink(A)) {
      LoopSoulDesc(A,Ai)
	many_global(true,Ai,all,Ai+1,A->end_soul_descendants());
    } else {
      LoopSoulDesc(A,Ai)
	many_global(is_sink(Ai),Ai,some,Ai+1,A->end_soul_descendants());
    }
#ifdef __GRAPE
  }
#endif
}
//------------------------------------------------------------------------------
void gravity_kernel::many_individual(cell* A) const
{
  if(al_sink(A)) {
    LoopSoulDesc(A,Ai)
      many_individual(true,Ai,all,Ai+1,A->end_soul_descendants());
  } else {
    LoopSoulDesc(A,Ai)
      many_individual(is_sink(Ai),Ai,some,Ai+1,A->end_soul_descendants());
  }
}
//------------------------------------------------------------------------------
// auxiliary functions.                                                         
//                                                                              
// If all nodes are supposed to be sources (the current state of the code),     
// these functions are trivial.                                                 
//------------------------------------------------------------------------------
template<class NA, class NB>
inline bool interacting(const NA*A, const NB*B) { return is_sink(A); }
//------------------------------------------------------------------------------
template<class NA, class NB>
inline real mutual_mass(const NA*A, const NB*B) { return mass(A)*mass(B); }    
//------------------------------------------------------------------------------
// inline real& c0(cell* C) { return    *(C->coeffs()); }
// inline ten1  c1(cell* C) { return ten1(C->coeffs()+N_C1); }
// inline ten2  c2(cell* C) { return ten2(C->coeffs()+N_C2); }
// inline ten3  c3(cell* C) { return ten3(C->coeffs()+N_C3); }
// inline ten4  c4(cell* C) { return ten4(C->coeffs()+N_C4); }
//------------------------------------------------------------------------------
//                                                                              
// Cell-Body interaction with approximated gravity                              
//                                                                              
// This routines performs a mutual cell-body interaction.                       
//                                                                              
//------------------------------------------------------------------------------
template<> inline
void gravity_kernel::grav<false>(vect &R, const real Rq, cell *A, soul *B) const
{
  GREENS_FUNCTION_3                                // define Dn (see kern.h)    
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    give_coeffs(A);                                //   guarantee that A has Ci 
    c0(A)+= D0;                                    //   add to C0 of A          
    c1(A).add_mul(R, D1);                          //   add to C1 of A          
    c2(A).add_quadrup(R,D1,D2);                    //   add to C2 of A          
    c3(A).add_quadrup(R,D2,D3);                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    register vect RQ;                              //   auxiliary vector        
    quad(A).inner_prod(R,RQ);                      //   R*Q_A                   
    register real tQ = trace(quad(A)), RQR=R*RQ;   //   tr(Q_A), R*Q*R          
    B->pot()-=          D0+ D1*tQ + D2*RQR;	   //   pot A->B                
    B->acc().sub_mul(R, D1+ D2*tQ + D3*RQR);       //   acc A->B (term in R)    
    RQ      *= D2;                                 //   R*Q_A*D2                
    B->acc().sub_twice(RQ);                        //   acc A->B (term in R*Q_A)
  }                                                // <                         
}
//------------------------------------------------------------------------------
template<> inline
void gravity_kernel::grav<true>(vect &R, const real Rq, cell *A, soul *B) const
{
  GREENS_FUNCTION_3                                // define Dn (see kern.h)    
  SYM3(C3); C3.quadrup(R,D2,D3);                   // pre-compute C3            
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    give_coeffs(A);                                //   guarantee that A has Ci 
    c0(A)+= D0;                                    //   add to C0 of A          
    c1(A).add_mul(R, D1);                          //   add to C1 of A          
    c2(A).add_quadrup(R,D1,D2);                    //   add to C2 of A          
    c3(A) += C3;                                   //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    register vect RQ;                              //   auxiliary vector        
    quad(A).inner_prod(R,RQ);                      //   R*Q_A                   
    register real tQ = trace(quad(A)), RQR=R*RQ;   //   tr(Q_A), R*Q*R          
    B->pot()-= C3.inner_prod(octo(A))              //   add to pot of B         
               +        D0+ D1*tQ + D2*RQR;	   //   pot A->B                
    B->acc().sub_mul(R, D1+ D2*tQ + D3*RQR);       //   acc A->B (term in R)    
    RQ      *= D2;                                 //   R*Q_A*D2                
    B->acc().sub_twice(RQ);                        //   acc A->B (term in R*Q_A)
  }                                                // <                         
}
//------------------------------------------------------------------------------
//                                                                              
// Cell-Cell interaction with approximated gravity                              
//                                                                              
// This routines performs a mutual cell-cell interaction.                       
// It's at the root of the whole code; about 50% of the CPU time is spent       
// in this routine and its (inlined) calls.                                     
//                                                                              
//------------------------------------------------------------------------------
template<> inline
void gravity_kernel::grav<false>(vect &R, const real Rq, cell* A, cell *B) const
{
  GREENS_FUNCTION_3                                // define Dn (see kern.h)    
  SYM2(C2); C2.quadrup(R,D1,D2);                   // pre-compute C2            
  SYM3(C3); C3.quadrup(R,D2,D3);                   // pre-compute C3            
  register vect RQ;                                // to hold R*Q               
  register real tQ, RQR;                           // to hold tr(Q), R*Q*R      
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    give_coeffs(A);                                //   guarantee that A has Ci 
    quad(B).inner_prod(R,RQ);                      //   R*Q_B                   
    tQ    = trace(quad(B));                        //   tr(Q_B)                 
    RQR   = R*RQ;                                  //   R*Q_B*R                 
    c0(A)+=          D0 + D1*tQ + D2*RQR;          //   add to C0 of A          
    c1(A).add_mul(R, D1 + D2*tQ + D3*RQR);         //   add to C1 of A (R)      
    RQ   *= D2;                                    //   R*Q_B*D2                
    c1(A).add_twice(RQ);                           //   add to C1 of A (R*Q_B)  
    c2(A)+= C2;                                    //   add to C2 of A          
    c3(A)+= C3;                                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    give_coeffs(B);                                //   guarantee that B has Ci 
    R.negate();                                    //   set R to -R             
    quad(A).inner_prod(R,RQ);                      //   R*Q_B                   
    tQ    = trace(quad(A));                        //   tr(Q_A)                 
    RQR   = R*RQ;                                  //   R*Q_A*R                 
    c0(B)+=          D0 + D1*tQ + D2*RQR;          //   add to C0 of B          
    c1(B).add_mul(R, D1 + D2*tQ + D3*RQR);         //   add to C1 of B (R)      
    RQ   *= D2;                                    //   R*Q_A*D2                
    c1(B).add_twice(RQ);                           //   add to C1 of B (R*Q_A)  
    c2(B)+= C2;                                    //   add to C2 of B          
    c3(B)-= C3;                                    //   add to C3 of B          
  }                                                // <                         
}
//------------------------------------------------------------------------------
template<> inline
void gravity_kernel::grav<true>(vect &R, const real Rq, cell* A, cell *B) const
{
  GREENS_FUNCTION_3                                // define Dn (see kern.h)    
  SYM2(C2); C2.quadrup(R,D1,D2);                   // pre-compute C2            
  SYM3(C3); C3.quadrup(R,D2,D3);                   // pre-compute C3            
  register vect RQ;                                // to hold R*Q               
  register real tQ, RQR;                           // to hold tr(Q), R*Q*R      
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    give_coeffs(A);                                //   guarantee that A has Ci 
    quad(B).inner_prod(R,RQ);                      //   R*Q_B                   
    tQ    = trace(quad(B));                        //   tr(Q_B)                 
    RQR   = R*RQ;                                  //   R*Q_B*R                 
    c0(A)+=          D0 + D1*tQ + D2*RQR           //   add to C0 of A          
                   - C3.inner_prod(octo(B));       //   add to pot of A         
    c1(A).add_mul(R, D1 + D2*tQ + D3*RQR);         //   add to C1 of A (R)      
    RQ   *= D2;                                    //   R*Q_B*D2                
    c1(A).add_twice(RQ);                           //   add to C1 of A (R*Q_B)  
    c2(A)+= C2;                                    //   add to C2 of A          
    c3(A)+= C3;                                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    give_coeffs(B);                                //   guarantee that B has Ci 
    R.negate();                                    //   set R to -R             
    quad(A).inner_prod(R,RQ);                      //   R*Q_B                   
    tQ    = trace(quad(A));                        //   tr(Q_A)                 
    RQR   = R*RQ;                                  //   R*Q_A*R                 
    c0(B)+=          D0 + D1*tQ + D2*RQR           //   add to C0 of B          
                   + C3.inner_prod(octo(A));       //   add to C0 of B         
    c1(B).add_mul(R, D1 + D2*tQ + D3*RQR);         //   add to C1 of B (R)      
    RQ   *= D2;                                    //   R*Q_A*D2                
    c1(B).add_twice(RQ);                           //   add to C1 of B (R*Q_A)  
    c2(B)+= C2;                                    //   add to C2 of B          
    c3(B)-= C3;                                    //   add to C3 of B          
  }                                                // <                         
}
//------------------------------------------------------------------------------
// auxiliary methods that are employed in deciding how to deal with a given     
// interaction between two nodes.                                               
//------------------------------------------------------------------------------
namespace nbdy {
  inline bool split_into_many(const cell   *a,     // I:  cell A                
			      const cell   *b,     // I:  cell B                
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-cell interaction can be done directly.
  {
//     return 4u < Ni && number(a)*number(b) < Ni;
    return number(a) < Ni && number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool split_into_many(const cell   *b,     // I:  cell                  
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-self or a cell-body interaction shall 
    // be done directly.                                                        
  {
    return number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *a,      // I:  cell A                
			     const cell   *b,      // I:  cell B                
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-cell interaction is well-separated.
  {
    return Rq > square(rcrit(a)+rcrit(b));
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *b,      // I:  cell                  
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-body interaction is well-separated.
  {
    return Rq > rcrit2(b);
  }
  //----------------------------------------------------------------------------
  inline bool are_twigs(const cell *a,             // I:   cell A               
			const cell *b)             // I:   cell B               
    // This routine figures out whether both cells are twig cells.              
  {
    return is_twig(a) && is_twig(b);
  }
  //----------------------------------------------------------------------------
  template<class NA, class NB>
  inline bool have_interaction(const NA*A, const NB*B)
    // This routines figures out whether any interaction is needed between two  
    // nodes A and B.                                                           
  { 
    return interacting(A,B) || interacting(B,A);
  }
}
//------------------------------------------------------------------------------
void gravity_kernel::interact(soul*A, soul*B) const
  // This routines performs an elementary body-body interaction.                
{
  if(!have_interaction(A,B)) return;               // IF no interaction > DONE <
  register vect R  = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(R),x,D0,ME;              // compute R^2               
  switch(SOFT) {                                   // what kind of softening? > 
  case individual:                                 // - individual softening >  
    BODYBODY_INDI(mutual_mass(A,B)) break;         //   compute (see gren.h) <  
  default:                                         // - global softening     >  
    BODYBODY_GLOB(mutual_mass(A,B)) break;         //   compute (see gren.h) <  
  }                                                // <                         
  if(interacting(A,B)){ A->pot()-=D0; A->acc()+=R;}// IF(B->A) >  add to A  <   
  if(interacting(B,A)){ B->pot()-=D0; B->acc()-=R;}// IF(A->B) >  add to B  <   
  STAT->record_BB();                               // record statistics         
}  
//==============================================================================
template<bool F> inline
bool gravity_kernel::mutual_global(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq)) {                       // IF(well separated)       >
    grav<F>(dX,Rq,A,B);                            //   interact                
    STAT->record_taylor_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) || split_into_many(A,N_POST[0])) { // IF(twig cell | many BBs) >
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
template<bool F> inline
bool gravity_kernel::mutual_individual(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_direct_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq)) {                       // IF(well separated)       >
    EPS = eps_mutual(A,B);                         //   set eps                 
    EPQ = EPS*EPS;                                 //   and eps^2               
    grav<F>(dX,Rq,A,B);                            //   interact                
    STAT->record_taylor_CB();                      //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) || split_into_many(A,N_POST[0])) { // IF(twig cell | many BBs) >
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_direct_CB();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, soul*B) const
{
  // This routines performs an elementary cell-body interaction.                
  if(!have_interaction(A,B)) return true;
//   if(FULL_POT)
    switch(SOFT) {
    case individual: return mutual_individual<false>(A,B);
    default:         return mutual_global    <false>(A,B);
    }
//   else
//     switch(SOFT) {
//     case individual: return mutual_individual<true> (A,B);
//     default:         return mutual_global    <true> (A,B);
//     }
}
//==============================================================================
template<bool F> inline
bool gravity_kernel::mutual_global(cell*A, cell*B) const
{
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq)) {                    // IF(well separated)      > 
    grav<F>(dX,Rq,A,B);                            //   interact                
    STAT->record_taylor_CC();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF(twig cells             
     split_into_many(A,B,N_POST[1])) {             //    OR make many BBs)    > 
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_direct_CC();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
template<bool F> inline
bool gravity_kernel::mutual_individual(cell*A, cell*B) const
{
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq)) {                    // IF(well separated)     >  
    EPS = eps_mutual(A,B);                         //   set eps                 
    EPQ = EPS*EPS;                                 //   and eps^2               
    grav<F>(dX,Rq,A,B);                            //   interact hard           
    STAT->record_taylor_CC();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF(twig cells             
     split_into_many(A,B,N_POST[1])) {             //    OR make many BBs)    > 
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_taylor_CC();                      //   record stats            
    return true;                                   //   DONE                    
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, cell*B) const
{
  // This routines performs an elementary cell-cell interaction.                
  if(!have_interaction(A,B)) return true;
//   if(FULL_POT)
    switch(SOFT) {
    case individual: return mutual_individual<false>(A,B);
    default:         return mutual_global    <false>(A,B);
    }
//   else
//     switch(SOFT) {
//     case individual: return mutual_individual<true> (A,B);
//     default:         return mutual_global    <true> (A,B);
//     }
}
//==============================================================================
bool gravity_kernel::interact(cell*A) const
{
  // This routines performs an elementary cell self-interaction.                
  if(!is_sink(A)) return true;
  switch(SOFT){
  case individual:
    if(is_twig(A) || split_into_many(A,N_PRE[2])) { 
      many_individual(A);
      STAT->record_direct_CS();
      return true;
    }
    return false;
  default:
    if(is_twig(A) || split_into_many(A,N_PRE[2])) {
      many_global(A);
      STAT->record_direct_CS();
      return true;
    }
    return false;
  }
}
//------------------------------------------------------------------------------
void gravity_kernel::direct_summation(cell*A) const
{
  switch(SOFT){
  case individual: return many_individual(A);
  default:         return many_global(A);
  }
}
//------------------------------------------------------------------------------
