
// code construction for tree re-building
// assume every thing from tree.cc




namespace fasttree {

  //----------------------------------------------------------------------------
  // we split the task into several steps
  // 1. we create a box-leaf tree, where the boxes have their child boxes in
  //    their OCT[] array, and all leafs that are still contained in the box
  //    in a linked list pointed to by box::LEAFS.
  //    We have to deal with the fact that some cells are not direct children
  //    of their parent cells, but grand-children (or later) if a zombie cell
  //    (or more) habe been eliminated.
  //    We also have to account for possible leakage out of the root cell.
  //    After this stage, the old cells and souls can be de-allocated
  // 2. we sort the leafs of the linked list back into the octants. This is 
  //    done recursively, such that we have ALL leafs that are in the current
  //    box at hand (including those of the box childreen) and can decide on
  //    NCRIT specific points. This stage is similar to tree_builder::add_leafs()
  // 3. we link the tree in very much the same way as tree_builder::link_cells().

  //----------------------------------------------------------------------------
  // class fasttree::tree_rebuilder<CELL>                                       
  //----------------------------------------------------------------------------
  template<class CELL>
  class tree_rebuilder :
    public builder_tools<CELL>
  {
    tree_rebuilder          (const tree_rebuilder&);// not implemented          
    tree_rebuilder&operator=(const tree_rebuilder&);// not implemented          
    //--------------------------------------------------------------------------
  private:
    tree<CELL> *OLDT;
    leaf_type  *L0, *LN;                           // begin/end of leafs        
    box_type   *P0;                                // root box                  
    uint        NL;                                // # leafs added so far      
    //--------------------------------------------------------------------------
    void make_zombie(const cell_type*par,          // I: parent cell            
		     const cell_type*kid,          // I: de-zombies child cell  
			   cell_type&zom)          // O: zombie cell            
    {
      register uint b = octant(center(par),center(kid));
      register real R = two*radius(kid);
      radius_(zom)    = R;
      center_(zom)[0] = (b&1)? center(par)[0]+R center(par)[0]-R;
      center_(zom)[1] = (b&2)? center(par)[1]+R center(par)[1]-R;
#if NDIM==3
      center_(zom)[2] = (b&4)? center(par)[2]+R center(par)[2]-R;
#endif
      nsouls_(zom)     = 0;
      csouls_(zom)     = 0;
      ncells_(zom)     = 1;
      ccells_(zom)     = kid;
      number_(zom)     = number(kid);
    }
    //--------------------------------------------------------------------------
    void drag_leafs_up(box_type  *P,               // I:   box                  
		       leaf_type*&List)            // I/O: leaf list to add to  
    {
      register leaf_type*Ln,*Li;                   // current leaf and next one 
      for(Li=P->LEAFS; Li; Li=Ln) {                // loop P's list of leafs  > 
	Ln=Li->NEXT;                               //   next leaf in this list  
	Li->NEXT=List;                             //   add to top of List      
	List    = Li;                              //   let List point to it    
      }                                            // <                         
      if(P->TYPE) {                                // IF (P has sub-boxes)    > 
	const node*NU=P->OCT+NSUB;                 //   end of P's octants      
	for(register node*N; N!=NU; ++N)           //   loop P's sub-boxes    > 
	  drag_leafs_up(PBOX(N),List);             //     RECURSIVE CALL      < 
      }                                            // <                         
    }
    //--------------------------------------------------------------------------
    box_type* stage_one(const real      &sR,       // I:   salf*radius of parent
			const cell_type* C,        // I:   the current tree cell
		              leaf_type*&Lf,       // I/O: free leafs           
		              leaf_type*&Lp)       // O:   list for loose leafs 
    {
      const real salf = 0.4;                       // 1/4 < salf < 1/2          
      // 1. deal with the possibility of C being a de-zombied descendant cell   
      if(radius(C) < hR) {                         // IF (cell is zombie)    >  
	cell_type Cz;                              //   get new cell and        
	make_zombie(C,c,Cz);                       //   make it to zombie       
	return stage_one(sR,Cz,Lf,Lo);             //   RECURSIVE call on zombie
      }                                            // <                         
      // 2. deal with child souls of cell                                       
      register box_type P;                         // get us a box              
      P.set_up(C);                                 // reset & copy cell cen&rad 
      register leaf_type*Li;                       // current leaf              
      LoopSoulKids(typename cell_type,C,s) {       // loop child souls        > 
	Li = Lf++;                                 //   get new leaf            
	Li->set_up(s);                             //   initiliaze it from soul 
	if(P.contains(Li)) {                       //   IF(in box)            > 
	  P.addleaf_to_list(Li);                   //     add to current box    
	  ++NL;                                    //     count leafs added     
	} else {                                   //   < ELSE >                
	  Li->NEXT=Lp;                             //     add to linked list    
	  Lp=Li;                                   //     of loose leafs        
	}                                          //   <                       
      }                                            // <                         
      // 3. deal with child cells of cell                                       
      if(has_cell_children(C)) {                   // IF cell has child cells  >
	register leaf_type *Lo=0,*Ln;              //   list: subs' loose leafs 
	register box_type  *sub;                   //   potential sub-box       
	register real salf_rad=salf*radius(C);     //   half radius of cell    >
	LoopCellKids(typename cell_type,C,c) {     //   loop child cells       >
	  sub = stage_one(salf_rad,c,Lf,Lo);       //     RECURSIVE call        
	  if(sub) P.addbox_to_octs(sub);           //     IF sub-box, put in oct
	}                                          //   <                       
	for(Li=Lo; Li; Li=Ln) {                    //   leafs lost from subs:  >
	  Ln = Li->NEXT;                           //     remember next in list 
	  if(P.contains(Li)) {                     //     IF (in box)          >
	    P.addleaf_to_list(Li);                 //       add to current box  
	    ++NL;                                  //       count leafs added   
	  } else {                                 //     < ELSE (not in box)  >
	    Li->NEXT=Lp;                           //       add to linked list  
	    Lp=Li;                                 //       of loose leafs      
	  }                                        //     <                     
	}                                          //   <                       
      }                                            // <                         
      if(P.NUMBER>1) return make_copybox(P,NL);    // IF(#leafs >1) return box  
      if(P.NUMBER==1) {                            // IF(#leafs==1)            >
	P.LEAFS->NEXT=Lp;                          //   add it to the list of   
	Lp=P.LEAFS;                                //   loose leafs             
      }                                            // <                         
      return 0;                                    // return NULL               
    }
    //--------------------------------------------------------------------------
    void extend_root(leaf_type*&Lp) const
    {
      register uint b = P0->octant(Lp);
      register box_type* Pn = new_box(NL);
      Pn->RADIUS = two * P0->RADIUS;
      Pn->pos()[0] = (b&1)? P0->center(0)+P0->RADIUS : P0->center(0) - P0->RADIUS;
      Pn->pos()[1] = (b&2)? P0->center(1)+P0->RADIUS : P0->center(1) - P0->RADIUS;
#if NDIM==3
      Pn->pos()[2] = (b&4)? P0->center(2)+P0->RADIUS : P0->center(2) - P0->RADIUS;
#endif
      Pn->addbox_to_octs(P0);
      register leaf_type *Lo=0, *Li, *Ln;
      for(Li=Lp; Li; Li=Li=Ln) {
	Ln = Li->NEXT;
	if(Pn->contains(Li)) {
	  Pn->addleaf_to_list(Li);
	  ++nl;
	} else {
	  Li->NEXT=Lo;
	  Lo = Li;
	}
      }
      P0 = Pn;
      if(Lo) extend_root(Lo)
    }
    //--------------------------------------------------------------------------
    void stage_two(      box_type *P,              // I: current box            
		   const uint      l)              // I: level                  
    {
      if(l>=DMAX) error("exceeding max depth");    // IF (level>DMAX) ERROR     
      const    node**NU=P->OCT+NSUB;               // end of P's octants        
      register node**N;                            // current sub-node          
      if(P->NUMBER <= NCRIT) {                     // IF (# leafs <= NCRIT)  >  
	if(P->TYPE)                                //   IF box has sub-boxes >  
	  for(N=P->OCT; N!=NU; ++N) if(*N)         //     loop P's sub-boxes >  
	    drag_leafs_up(PBOX(*N),P->LEAFS);      //       get leafs up     < <
      } else {                                     // < ELSE >                  
	register int        b;                     //   current octant          
	register leaf_type *Li,*Ln,*Lo;            //   current, next & old leaf
	for(Li=P->LEAFS; Li; Li=Ln) {              //   loop leafs in P's list >
 	  Ln = Li->NEXT;                           //     next leaf in list     
	  b  = P->octant(Li);                      //     leaf's octant         
	  N  = P->OCT+b;                           //     pointer to octant     
	  if((*oc)==0) {                           //     IF octant empty >     
	    *N = Li;                               //       assign leaf to it   
	    P->mark_as_leaf(b);                    //       mark octant as leaf 
	    ++NL;                                  //       count leafs added   
	  } else if(P->marked_as_leaf(b)) {        //     < ELSE (octant=leaf) >
	    Lo = PLEAF(*N);                        //       get old leaf        
	    P->mark_as_box(b);                     //       mark octant as box  
	    P = make_subbox(P,b,Lo,NL);            //       create sub-box      
	    *N = P;                                //       assign sub-box to oc
	  } else                                   //     < ELSE (octant=box) > 
	    PBOX(*N)->addleaf_to_list(Li);         //       add leaf to list    
	}                                          //   <                       
	P->LEAFS=0;                                //   empty P's leaf list     
	for(b=0,N=P->OCT; N!=NU; ++b,++N) if(*N) { //   loop non-empty octants >
	  if(P->marked_as_box(b))                  //     IF (box)             >
	    stage_two(PBOX(*N),l+1);               //       RECURSIVE call      
	  else {                                   //     < ELSE (leaf)        >
	    PLEAF(*N)->NEXT = P->LEAFS;            //       add to list         
	    P->LEAFS = PLEAF(*N);                  //       of P's leafs        
	    *N = 0;                                //       delete octant entry 
	  }                                        //     <                     
	}                                          //   <                       
      }                                            // <                         
    }
    //--------------------------------------------------------------------------
    template<bool U>
    uint stage_three(
		     box_type * P,                 // current box               
		     cell_type* C,                 // current cell              
		     cell_type*&Cf,                // free cells                
		     soul_type*&Sf) const          // free souls                
    {

      register uint dep=0;                         // depth of cell             
      radius_(C) = P->RADIUS;                      // copy radius               
      center_(C) = P->center();                    // copy center               
      number_(C) = P->NUMBER;                      // copy number               
      csouls_(C) = Sf;                             // set cell: soul kids       
      nsouls_(C) = 0;                              // reset cell: # soul kids   
      if(P->TYPE==0) {                             // IF(box==twig)         >   
	ccells_(C) = 0;                            //   set cell: sub-cells     
	ncells_(C) = 0;                            //   set cell: # sub-cells   
	register leaf_type*Li=P->LEAFS;            //   sub-leaf pointer        
	for(; Li; Li=Li->NEXT) {                   //   loop sub-leafs      >   
	  if(U) Li->set_soul_B(Sf++);              //     set soul              
	  else  Li->set_soul_E(Sf++);              //     set soul              
	  nsouls_(C)++;                            //     increment # sub-souls 
	}                                          //   <                       
      } else {                                     // < ELSE (box==branch)  >   
	register uint   i;                         //   index to octant         
	register node **N;                         //   pointer to current sub  
	de_zombie(P);                              //   de-zombie               
	register uint de,nsub=P->N_subboxes();     //   number of sub-boxes     
	if(nsub) {                                 //   IF sub-boxes         >  
	  register box_type *Pi;                   //     sub-box pointer       
	  register cell_type*Ci=Cf;                //     remember free cells   
	  ncells_(C) = nsub;                       //     set cell: sub-cells   
	  ccells_(C) = Cf;                         //     set cell: sub-cells   
	  Cf += nsub;                              //     reserve nsub cells    
	  for(i=0,N=P->OCT; i!=NSUB; ++i,++N)      //     loop octants     >    
	    if(*N) {                               //       IF sub-box     >    
	      Pi = PBOX(*N);                       //         sub-box           
	      de = link_cells<U>(Pi,Ci++,Cf,Sf);   //         recursive call    
	      if(de>dep) dep=de;                   //         update depth      
	    }                                      //     < <                   
	} else {                                   //   < ELSE (no sub-boxes) > 
	  ncells_(C) = 0;                          //     set cell: sub-cells   
	  ccells_(C) = 0;                          //     set cell: sub-cells   
	}                                          //   <                       
      }                                            // <                         
      dep++;                                       // increment depth           
      return dep;                                  // return cells depth        
    }
    //--------------------------------------------------------------------------
  public:
    bool stage_one()
    {
      L0 = new leaf_type[OLDT->N_souls()];
      NL = 0;
      register leaf_type *Lp=0;
      P0 = stage_one(0.4*radius(OLDT->root()), OLDT->root(),L0,Lp);
      if(!P0) return false;
      if(Lp) extend_root(Lp);
      return true;
    }
    //--------------------------------------------------------------------------
    void stage_two()
    {
      register uint level=0;
      stage_two(P0,level);
    }
    //--------------------------------------------------------------------------
    void stage_three(cell_type*&Cf, soul_type*&Lf)
    {
      cell_type *C0=Cf++;
      if(OLDT->use_bodies()) stage_three<true> (P0,C0,Cf,Lf);
      else                   stage_three<falce>(P0,C0,Cf,Lf);
    }
    //--------------------------------------------------------------------------
    tree_rebuilder(const uint nc, const uint dm, const tree<CELL>* oldtree)
      : builder_tools(nc,dm), OLDT(oldtree)
    {
      BM = new manager<box_type>((oldtree->N_cells()*11)/10,
				  oldtree->N_souls());
    }
    //--------------------------------------------------------------------------
    tree_rebuilder() {
      delete BM;
    }
  };
}
























