//-----------------------------------------------------------------------------+
//                                                                             |
// tayl.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <nbdy/tayl.h>
#include <nbdy/soul.h>
#include <nbdy/cell.h>
#ifdef __GRAPE
#  include <nbdy/graa.h>
#endif

#include <iomanip>


using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// class nbdy::TaylorExpansion                                                  
////////////////////////////////////////////////////////////////////////////////
inline
void TaylorExpansion::shift_by(real*&T, const vect& X) const {
  //----------------------------------------------------------------------------
  // computational effort:                                                      
  //                  p   |  1    2    3    4    5    6    7                    
  //                 -----+----------------------------------                   
  //                  *   |  3   18   51  114  222  393  648                    
  //                  +/- |  1   12   35   78  151  266  437                    
  //----------------------------------------------------------------------------
  if(X!=zero && c0(T) != zero ) {                   // 1st order                
    real         &C0(*T);                           //   C0                     
    register ten1 C1(T+N_C1);                       //   C1                     
    C0 += X * C1;                                   //   1st order shift: C0    
  if(PO > 1) {                                      // 2nd order                
    register vect X2 = half*X;                      //   X/2                    
    register vect F1;                               //   auxiliary Tensor_1     
    register ten2 C2(T+N_C2);                       //   C2                     
    C2.inner_prod(X,F1);                            //   F1 = X * C2            
    C1 += F1;                                       //   2nd order shift: C1    
    C0 += X2 * F1;                                  //   2nd order shift: C0    
  if(PO > 2) {                                      // 3rd order                
    register vect X3 = third*X;                     //   X/3                    
    SYM2(F2);                                       //   auxiliary Tensor_2     
    register ten3 C3(T+N_C3);                       //   C3                     
    C3.inner_prod(X,F2);                            //   F2 = X * C3            
    C2 += F2;                                       //   3rd order shift: C2    
    F2.inner_prod(X2,F1);                           //   F1 = X^2*C3 / 2        
    C1 += F1;                                       //   3rd order shift: C1    
    C0 += X3 * F1;                                  //   3rd order shift: C0    
  if(PO > 3) {                                      // 4th order                
    register vect X4 = quarter*X;                   //   X/4                    
    SYM3(F3);                                       //   auxiliary Tensor_3     
    register ten4 C4(T+N_C4);                       //   C4                     
    C4.inner_prod(X,F3);                            //   F3 = X * C4            
    C3 += F3;                                       //   3rd order shift: C3    
    F3.inner_prod(X2,F2);                           //   F2 = X^2*C4 / 2        
    C2 += F2;                                       //   3rd order shift: C2    
    F2.inner_prod(X3,F1);                           //   F1 = X^3*C4 / 6        
    C1 += F1;                                       //   3rd order shift: C1    
    C0 += X4 * F1;                                  //   3rd order shift: C0    
  if(PO > 4) {                                      // 5th order                
    error(" expansion order > 4 currently not supported");
//     register vect X5 = fifth*X;                     //   X/5                    
//     SYM4(F4);                                       //   auxiliary Tensor_4     
//     register ten5 C5(T+N_C5);                       //   C5                     
//     C5.inner_prod(X,F4);                            //   F4 = X * C5            
//     C4 += F4;                                       //   3rd order shift: C4    
//     F4.inner_prod(X2,F3);                           //   F3 = X^2*C5 / 2        
//     C3 += F3;                                       //   3rd order shift: C3    
//     F3.inner_prod(X3,F2);                           //   F2 = X^3*C5 / 6        
//     C2 += F2;                                       //   3rd order shift: C2    
//     F2.inner_prod(X4,F1);                           //   F1 = X^4*C5 / 24       
//     C1 += F1;                                       //   3rd order shift: C1    
//     C0 += X5 * F1;                                  //   3rd order shift: C0    
  } } } } }
}
//------------------------------------------------------------------------------
inline void TaylorExpansion::extract_grav(const TaylorSeries&T, soul*S) const
{
  //----------------------------------------------------------------------------
  // computational effort:                                                      
  //                  p   |  1    2    3    4    5    6    7                    
  //                 -----+----------------------------------                   
  //                  *   |  3   15   43  114  222  393  648                    
  //                  +/- |  5   15   37   78  151  266  437                    
  //----------------------------------------------------------------------------
  register vect X = cofm(S)-T.X;                    // expansion argument       
  register vect A = c1(T);                          //   1st order acceleration 
  register real P = c0(T) + X * c1(T);              //   1st order potential    
  if(PO > 1) {                                      // 2nd order                
    register vect F;                                //   auxiliary vector       
    c2(T).inner_prod(X,F);                          //   F = X * C2             
    A += F;                                         //   2nd order acceleration 
    P += half * (X*F);                              //   2nd order potential    
  if(PO > 2) {                                      // 3rd order                
    const real i3= third*half;                      //   1/3!                   
    SYM2(X2); X2.outer_prod(X);                     //   X2ij = Xi*Xj           
    c3(T).inner_prod(X2,F);                         //   F = X^2*C3             
    A += half * F;                                  //   3rd order acceleration 
    P += i3 * (X*F);                                //   3rd order potential    
  if(PO > 3) {                                      // 4th order                
    const real i4= quarter*i3;                      //   1/4!                   
    SYM3(X3); X3.outer_prod(X2,X);                  //   X3ijk = Xi*Xj*Xk       
    c4(T).inner_prod(X3,F);                         //   F = X^3*C4             
    A += i3 * F;                                    //   4th order acceleration 
    P += i4 * (X*F);                                //   4th order potential    
  if(PO > 4) {                                      // 5th order                
    error(" expansion order > 4 currently not supported");
//     const real i5= fifth*i4;                        //   1/5!                   
//     SYM4(X4); X4.outer_prod(X3,X);                  //   X4ijkl = Xi*Xj*Xk*Xl   
//     c5(T).inner_prod(X4,F);                         //   F = X^4*C5             
//     A += i4 * F;                                    //   5th order acceleration 
//     P += i5 * (X*F);                                //   5th order potential    
  } } } }
  S->acc() += A;                                    // add to souls acceleration
  S->pot() -= P;                                    // add to souls potential   
}
//------------------------------------------------------------------------------
inline
void TaylorExpansion::shift_to(TaylorSeries&T, const vect&X) const {
  shift_by(T.C,X-T.X);
  T.X = X;
}
//------------------------------------------------------------------------------
void TaylorExpansion::shift_and_add(TaylorSeries&T,
				    const vect  &x,
				    const real  *t,
				    const real  &f) const {
  if(t && t[0]) {
    shift_to(T,x);
    for(register int i=0; i!=NG; ++i) T.C[i] += t[i] * f;
  }
}
//------------------------------------------------------------------------------
inline
void TaylorExpansion::evaluate_grav(cell* C, const TaylorSeries&T
#ifdef __GRAPE
				    ,const grape_arrays*A
#endif
				    ) const
{
  TaylorSeries G(T.X,make_coeffs(T.C));             // G = copy of T            
  shift_and_add(G,cofm(C),coeffs(C),imass(C));      // shift G; G+=T_C          
  take_coeffs(C);                                   // free memory for T_C      
  LoopSoulKids(cell,C,s) if(is_sink(s)) {           // loop C's sink soul kids  
#ifdef __GRAPE
    if(A) {                                         //   IF(GRAPE) >            
      if(pot(s)) s->normalize_grav();               //     pot,acc/=mass        
      register int i=A->index(s);                   //     index of soul        
      s->acc() += A->acc(i);                        //     add pot from GRAPE   
      s->pot() += A->pot(i);                        //     add acc from GRAPE   
    } else                                          //   < ELSE >               
#endif
      s->normalize_grav();                          //     pot,acc/=mass   <    
    extract_grav(G,s);                              //   add pot,acc due to G   
  }                                                 // <                        
  LoopCellKids(cell,C,c) if(is_sink(c))             // loop C's sink cell kids  
#ifdef __GRAPE
    evaluate_grav(c,G,A);                           //   recursive call         
#else
    evaluate_grav(c,G);                             //   recursive call         
#endif
  take_coeffs(G.C);                                 // free memory for G        
}
//------------------------------------------------------------------------------
#ifdef __GRAPE
void TaylorExpansion::evaluate_grav(cell*C, const grape_arrays*A) const
#else
void TaylorExpansion::evaluate_grav(cell*C) const
#endif
{
  if(NC > MAXNC) MAXNC = NC;
  TaylorSeries G(cofm(C),make_coeffs());            // G = empty Taylor series  
#ifdef __GRAPE
  evaluate_grav (C,G,A);                            // start recursion          
#else
  evaluate_grav (C,G);                              // start recursion          
#endif
  take_coeffs(G.C);                                 // free memory for G        
}
////////////////////////////////////////////////////////////////////////////////
