// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tcdm.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2002                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <tcdm.h>
#include <util/Pi.h>
#include <util/ionl.h>
#include <util/tupl.h>
#include <util/nums.h>
#include <util/spln.h>
#include <util/Math.h>
#include <fstream>
#include <iomanip>
using namespace util;
using namespace nbdy;

#define SECH_REDUC 1                                // sech(r/b) truncation     

//==============================================================================
// auxiliary functions                                                          
//==============================================================================
#ifdef SECH_REDUC
inline double reduc_fac(                            // density reduction factor 
			const double x)             // I: x = r/b               
{ 
  register double ex = exp(-x);
  return 2.*ex/(1+ex*ex);
}
//------------------------------------------------------------------------------
inline double reduc_fac(                            // density reduction factor 
			const double x,             // I: x = r/b               
		              double&d1,            // O: df/dx                 
		              double&d2)            // O: d^2f/dx^2             
{ 
  register double ex = exp(-x), e2x=ex*ex, fac=1./(1.+e2x);
  d1=-2.*(1.-e2x)*ex*square(fac);
  d2= 2*(ex*(1+e2x*(e2x-6)))*cube(fac);
  return 2.*ex*fac;
}
#else
//------------------------------------------------------------------------------
inline double reduc_fac(                            // density reduction factor 
 			const double x)             // I: x = r/b               
{ return exp(-x); }
//------------------------------------------------------------------------------
inline double reduc_fac(                            // density reduction factor 
			const double x,             // I: x = r/b               
		              double&d1,            // O: df/dx                 
		              double&d2)            // O: d^2f/dx^2             
{ 
  register double ex = exp(-x);
  d1=-ex;
  d2= ex;
  return ex;
}
#endif
//------------------------------------------------------------------------------
static double AB_;                                  // a/b for global functions 
static double ALF;                                  // alf for global functions 
//------------------------------------------------------------------------------
inline double rho_cdm(                              // CDM density              
		      const double x,               // I: x = r/a 
		      const double alf)             // I: alpha                 
{
  if     (alf == 0)
    return 1./cube(x+1);
  else if(alf == 1)
    return 1./(x*square(x+1));
  else if(alf == 2)
    return 1./(square(x)*(x+1));
  register double y=x/(x+1);
  return (1.-y)/pow(y,alf);
}
//------------------------------------------------------------------------------
inline double rho_cdm(                              // CDM density              
		      const double x,               // I: x = r/a               
		            double&d1,              // O: drho/dx               
		            double&d2,              // O: d^2rho/dx^2           
		      const double alf)             // I: alpha                 
{
  register double
    ix = 1/x,                                       // 1/x                      
    y1 = 1/(x+1),                                   // 1-y = 1/(x+1)            
    y  = x*y1,                                      // y   = x/(x+1)            
    ya = pow(y,alf),                                // y^alfa                   
    rh = cube(y1)/ya;                               // density                  
  d1   =-y1   *rh*( 3+alf*ix);
  d2   = y1*y1*rh*(12+alf*ix*(8+(alf+1)*ix));
  return rh;
}
//------------------------------------------------------------------------------
inline double rho_tcdm(                             // reduced density          
		       const double x)              // I: x = r/a               
{ 
  return reduc_fac(x*AB_) * rho_cdm(x,ALF);         // using global a/b        
}
//------------------------------------------------------------------------------
inline double rho_tcdm(                             // reduced density          
		       const double x,              // I: x = r/a               
		             double&d1,             // O: drho/dx               
		             double&d2)             // O: d^2rho/dx^2           
{
  register double 
    rf1,rf2,rf=reduc_fac(x*AB_,rf1,rf2),
    rh1,rh2,rh=rho_cdm  (x,rh1,rh2,ALF);
  rf1*= AB_;
  rf2*= square(AB_);
  d1 = rf1*rh + rf*rh1;
  d2 = rf2*rh + 2*rf1*rh1 + rf*rh2;
  return rf*rh;
}
//------------------------------------------------------------------------------
inline double dM(const double lr, const double M)   // dM/dln r                 
{
  register double r = exp(lr);                      // ln r as independent var
  if     (ALF==0)
    return reduc_fac(r*AB_) * cube(r/(r+1));
  else if(ALF==1)
    return reduc_fac(r*AB_) * square(r/(r+1));
  else if(ALF==2)
    return reduc_fac(r*AB_) * r/(r+1);
  else
    return reduc_fac(r*AB_) * pow(r/(r+1),3-ALF);
}
//------------------------------------------------------------------------------
static spline<double> *SPLINE;                      // global spline            
//------------------------------------------------------------------------------
inline double glob_spline(const double M) { return (*SPLINE)(M); }
//------------------------------------------------------------------------------
inline tupel<double,2> dP(const double lr, const tupel<double,2> M)
{
  register double          r = exp(lr);
  register tupel<double,2> F;
  F[0] =-(*SPLINE)(lr) / r;                  // dP             = M/r * dln r    
  F[1] = rho_tcdm(r) * F[0];                 // d(rho*sigma^2) = rho * M/r dln r
  return F;
}
//------------------------------------------------------------------------------
inline double special(const double x) {
  // int(exp(-t)/t^5,t=x...oo)                                                  
  return ( (6+x*(x*(1-x)-2))*exp(-x)/pow(x,4) + En(1,x) ) / 24.;
}
//------------------------------------------------------------------------------
inline double Fpow(const double x, const double p)
{
  register double p1=p+1;
  if(p1) return pow(x,p1)/p1;
  else   return log(x);
}
//==============================================================================
// set up:  radial grid, cumulative mass, potential and velocity dispersion     
//==============================================================================
inline void tcdm_model::setup(const double V)
{
  ab  = a/b;
  AB_ = ab;
  ALF = A;
  register double A3=3-ALF;
  // 1. initialize grid, put density on grid
  register double 
    ba   = b/a,
    rmin = 0.001*min(ba,1.),
    rmax = 100*max(ba,1.);
  n1 = int(200*log10(rmax/rmin));
  n  = 1+n1;
  register double 
    dlr  = log(rmax/rmin)/double(n1),
    *rh  = new double[n];
  lr     = new double[n];
  r      = new double[n];
  r [0]  = rmin;
  lr[0]  = log(rmin);
  rh[0]  = rho_tcdm(rmin);
  for(register int i=1; i!=n; ++i) {
    lr[i] = lr[0] + i * dlr;
    r [i] = exp(lr[i]);
    rh[i] = rho_tcdm(r[i]);
  }
  // 2. cumulative mass, maximum v_c, set scale factors
  m    = new double[n];
  register double M = pow(rmin,A3)/A3*(1-square(A3)/(4-ALF)*rmin);
  m[0] = FPi * M;
  vqm  = m[0] / r[0];
  for(register int i=1; i!=n; ++i) {
    M    = rk4(M,lr[i-1],dlr,dM);
    m[i] = FPi * M;
    if(m[i] > vqm * r[i]) vqm = m[i] / r[i];
  }
  ca  = square(V) / vqm;                  // c/a  : scale for pot, vc^2, sigma^2
  c   = a * ca;                           // c    : scale for mass              
  cac = ca / square(a);                   // c/a^3: scale for density           
  // 3. potential & isotropic velocity dispersion
  double s0,sN;
  s0    = FPi * pow(r[ 0],3) * rh[ 0],
  sN    = FPi * pow(r[n1],3) * rh[n1];
  SPLINE= new spline<double>(n,lr,m,&s0,&sN);
  ps    = new double[n];
  sq    = new double[n];
  register double rhsq = m[n1]*pow(ab,4)*2*special(r[n1]*ab), err;
  register tupel<double,2> y;
  ps[n1] = m[n1] / r[n1];
  sq[n1] = rh[n1]? rhsq / rh[n1] : 0;
  y[0] = ps[n1];
  y[1] = rhsq;
  for(register int i=n-2; i!=-1; --i) {
    y     = rk4(y,lr[i+1],-dlr,dP);
    ps[i] = y[0];
    sq[i] = rh[i]? y[1] / rh[i] : 0.;
  }
  if(A<2) ps0 = ps[0] + m[0]/(r[0]*(2-A));
  else    ps0 = 0;
  sq0 = sq[0]/pow(r[0],A) + FPi/(3-A)*Fpow(r[0],1-A-A);
  delete   SPLINE;
  delete[] rh;
  // 4. total energy
  if(A < 2.5) {
    for(nm=0; nm!=n1; ++nm) if(m[nm+1] == m[nm]) break;
    nm++;
    s0     =-1./(r[0]*A3);
    SPLINE = new spline<double>(nm,m,ps,&s0,0);
    etot   = -0.25 * (m[0]*(ps[0]+m[0]/((5-A+A)*r[0])) +
		      qbulir(glob_spline,m[0],m[nm-1],1.e-8,&err,0) );
  } else
    etot = 0;
  // 5. distibution function of isotropic model
  setfE();
}
//==============================================================================
static double E_,M_,aM_;                         // Eps, Mtot, Mtot*a/b         
inline double d2rhodpsi2_of_q(const double q)
{
  register double psi = E_*(1.-q*q);             // Psi = Eps * (1 - q^2)       
  if     (psi <= 0.)
    return 0.;
  else if(psi < SPLINE->last_X()) {
    register double aMps=aM_/psi;
    return 2*(6*psi+aM_*(4+aMps))*exp(-aMps)/cube(M_);
  } else
    return (*SPLINE)(psi);
}
//------------------------------------------------------------------------------
inline double intfE(const double z)
{
  register double f=cube(z);
  return f*d2rhodpsi2_of_q(z*f);
}
//------------------------------------------------------------------------------
void tcdm_model::setfE()
{
  // 1. put d^2rho/dpsi^2 on grid
  double *in = new double[n];
  register double rh,rh1,rh2,ps1,ps2;
  for(register int i=0; i!=n; ++i) {
    rh    = rho_tcdm(r[i],rh1,rh2);               // rho, rho', rho''           
    ps1   =-m[i]/square(r[i]);                    // psi'                       
    ps2   =-2*ps1/r[i]-FPi*rh;                    // psi''                      
    in[i] = (rh2*ps1 - rh1*ps2)/cube(ps1);        // d^2rho/dpsi^2              
  }
  // 2. compute isotropic f(E) on grid
  register double s0, err;
  if     (A <2) s0 =-(4-A)*in[0]*r[0]/m[0];
  else if(A==2) s0 =-in[0]/TPi;
  else          s0 = (4-A)/(A-2)*in[0]/ps[0];
  SPLINE = new spline<double>(n,ps,in,&s0);

  M_     = m[n1];
  aM_    = ab*m[n1];
  lf     = new double[n];
  for(register int i=0; i!=n; ++i) {
    E_    = ps[i];
    lf[i] = log(4*sqrt(0.5*E_)/square(Pi)*qbulir(intfE,0.,1.,1.e-8,&err,0));
    if(err > 1.e-3) {
      cerr<<"tcdm_model: very inaccurate integration for f(E)\n";
      exit(1);
    }
  }
  delete SPLINE;
  delete[] in;
  fac    = c / pow(c*a,1.5);
  ln_fac = log(fac);
}
//==============================================================================
double tcdm_model::random(double(*R)(),
			  double&rad,
			  double&vel) const
{
  const    double tt = 2./3.;
  register double x,psi,mr=R()*m[n1];
  if(mr<m[0]) {
    x   = r[0]*pow(mr/m[0],1./(3-A));
    if     (A <2) psi = ps[0]+m[0]/(r[0]*(2-A))*(1-pow(x/r[0],2-A));
    else if(A==2) psi = FPi * log(x);
    else          psi = m[0]/(r[0]*(A-2))*pow(x/r[0],2-A);
  } else {
    x   = polev(mr,m,r,nm);
    psi = polev(mr,m,ps,nm);
  }
  rad = x * a;
  register double vq, fv, f0=fiso(psi);
  do {
    vq = pow(R(),tt) * 2 * psi;
    fv = fiso(psi-0.5*vq);
  } while(fv < R()*f0);
  vel = sqrt(ca*vq);
  return -ca*psi;
}
//==============================================================================
tcdm_model::tcdm_model(const double al,           // density exponent           
		       const double a_,           // break radius               
		       const double b_,           // sech^2 truncation radius   
		       const double v_) :         // maximal v_circ             
  A    ( al ),			 
  a    ( a_ ),
  b    ( b_ ),
  ln_a ( log(a) )
{ 
  setup(v_);
}
//------------------------------------------------------------------------------
tcdm_model::~tcdm_model()
{
  if(r)  delete[] r;
  if(lr) delete[] lr;
  if(m)  delete[] m;
  if(ps) delete[] ps;
  if(sq) delete[] sq;
  if(lf) delete[] lf;
}
//------------------------------------------------------------------------------
void tcdm_model::set_vc_max(const double&V)
{
  register double vqm=m[n/2]/r[n/2];
  for(register int i=0; i!=n; ++i)
    if(m[i] > vqm * r[i]) vqm = m[i]/r[i];
  ca  = square(V) / vqm;                  // c/a  : scale for pot, vc^2, sigma^2
  c   = a * ca;                           // c    : scale for mass              
  cac = ca / square(a);                   // c/a^3: scale for density           
}
//==============================================================================
inline
double tcdm_model::rhox(const double&x) const {
  return reduc_fac(x*ab) * rho_cdm(x,A);
}
//------------------------------------------------------------------------------
double tcdm_model::rho(const double&rad) const {
  return cac * rhox(rad/a);
}
//------------------------------------------------------------------------------
inline
double tcdm_model::psix(const double&x) const {
  if(x<= 0.)   return ps0;
  if(x< r[0] ) {
    if(A< 2.) return  ps[0]+m[0]/(r[0]*(2-A))*(1-pow(x/r[0],2-A));
    if(A==2.) return  FPi * log(x);
    else      return  m[0]/(r[0]*(A-2))*pow(x/r[0],2-A);
  }
  if(x> r[n1]) return m[n1]/x;
  else         return polev(log(x),lr,ps,n);
}
//------------------------------------------------------------------------------
double tcdm_model::pot(const double&rad) const {
  return -ca*psix(rad/a);
}
//------------------------------------------------------------------------------
inline
double tcdm_model::cumx(const double&x) const {
  if(x<= 0.)  return 0.;
  if(x<r[0] ) return m[0]*pow(x/r[0],3-A);
  if(x>r[n1]) return m[n1];
  else        return polev(log(x),lr,m,n);
}
//------------------------------------------------------------------------------
double tcdm_model::cum(const double&rad) const {
  return c*cumx(rad/a);
}
//------------------------------------------------------------------------------
double tcdm_model::vcq(const double&rad) const {
  if(rad <= 0.) return 0.;
  return cum(rad)/rad;
}
//------------------------------------------------------------------------------
double tcdm_model::siq(const double&rad) const {
  if(rad <= 0.) return 0.;
  register double x=rad/a;
  if(x<r[0] ) return ca*pow(x,A)*(sq0-FPi/(3-A)*Fpow(x,1-A-A));
  if(x>r[n1]) {
    register double rh=rho(rad);
    return rh? ca*cac*m[n1]*pow(ab,4)*2*special(x*ab)/rh : 0.;
  }
  return ca*polev(log(x),lr,sq,n);
}
//------------------------------------------------------------------------------
inline
double tcdm_model::fiso(const double&e) const {
  if(e <= ps[n] || A>2 && e>ps0) return 0.;
  if(e >ps[0]) {
    if(A==0) return exp(lf[0]) * (ps0-ps[0]) / (ps0-e);
    if(A <2) return exp(lf[0] - 0.5*(6-A)/(2-A)*log((ps0-e)/(ps0-ps[0])));
    if(A==2) return exp(lf[0] + 2*(e-ps[0]));
             return exp(lf[0] + 0.5*(6-A)/(A-2)*log(e/ps[0]));
  }
  return exp(polev(e,ps,lf,n) );
}
//------------------------------------------------------------------------------
double tcdm_model::f_E(const double&E) const {
  return fac * fiso(-E/ca);
}
////////////////////////////////////////////////////////////////////////////////
#if 0
int main(int argc, char* argv[])
{
  register double a = (argc>1)? atof(argv[1]) : 1.;
  register double b = (argc>2)? atof(argv[2]) : 10., r;
  tcdm_model CDM(a,1.,b,1.);
  cout << "\n" 
       << " E_tot =" << CDM.total_energy() << "\n\n";
  r = 0.5 * CDM.rad_grid(0);
  cout << "     "
       << std::setw(10) << r  << " "
       << std::setw(13) << CDM.vcq(r) * r << " "
       << std::setw(13) <<-CDM.pot(r) << " "
       << std::setw(13) << CDM.vcq(r) << " "
       << std::setw(13) << CDM.siq(r) << " "
       << std::setw(13) << log(CDM.f_E(CDM.pot(r))) << "\n";
  for(register int i=0; i!=CDM.N_grid(); ++i)
    cout << std::setw(4)  << i <<" "
	 << std::setw(10) << CDM.rad_grid(i) << " "
	 << std::setw(13) << CDM.cum_grid(i) << " "
	 << std::setw(13) << CDM.psi_grid(i) << " "
	 << std::setw(13) << CDM.vcq_grid(i) << " "
	 << std::setw(13) << CDM.siq_grid(i) << " "
	 << std::setw(13) << CDM.lnf_grid(i) << "\n";
  
  r = 2 * CDM.rad_grid(CDM.N_grid()-1);
  cout << "     "
       << std::setw(10) << r  << " "
       << std::setw(13) << CDM.cum(r) << " "
       << std::setw(13) <<-CDM.pot(r) << " "
       << std::setw(13) << CDM.vcq(r) << " "
       << std::setw(13) << CDM.siq(r) << " "
       << std::setw(13) << log(CDM.f_E(CDM.pot(r))) << "\n";
}
#endif
