// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// AnalysisGC.cc                                                               |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// Analysis of GC destruction run with YancNemo                                |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0.0   06/06/2002  WD created employing old program GCSimNemo            |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile AnalysisGC
#endif

#include <iostream>                                   // C++ I/O                
#include <iomanip>                                    // C++ I/O manipulators   
#include <fstream>                                    // C++ file I/O           
using std::cerr;
using std::ostream;
using std::ofstream;
using std::setprecision;
#include <body.h>                                     // bodies                 
#include <nbdy/nmio.h>                                // my nemo snapshot I/O   
#include <util/ionl.h>                                // I/O helper             

extern "C" {
#include <nemo.h>                                     // NEMO basics            
}
//------------------------------------------------------------------------------
// some auxiliary methods that are unique to this main                          
//------------------------------------------------------------------------------
namespace nbdy {
  class GCOrbit {
  private:
            real RQ;
    mutable vect X0,V0;
    mutable real S0,Sb;
    mutable uint N0,Nb;
    //--------------------------------------------------------------------------
  public:
    GCOrbit(const real r) : RQ (r*r) {}
    //--------------------------------------------------------------------------
    void set(bodies *BB, const real time) {
      register double               M=0., iM;
      register nbdy::vector<double> X=0.;
      // determine c-of-m pos of all bodies unmarked                            
      LoopBodies(bodies,BB,Bi) if(!is_marked(Bi)) {
	X+= mass(Bi) * pos(Bi);
	M+= mass(Bi);
      }
      iM = M>0.? 1./M : 1.;
      X0 = X*iM;
      // mark/unmark bodies if they are within/outside of r_cut of x0           
      N0 = 0u;
      if(RQ > 0.) {
	X = M = 0.;
	LoopBodies(bodies,BB,Bi)                 // loop all bodies           > 
	  if(dist_sq(pos(Bi),X0) > RQ) {         //   IF(outside of sphere)   > 
	    if(!is_marked(Bi)) {                 //     IF(outside 1st time) >  
	      Bi.mark();                         //       mark body             
	      Bi.aux() = time;                   //       remember this time    
	    }                                    //     <                       
	  } else {                               //   < ELSE (inside) >         
	    if(is_marked(Bi)) Bi.unmark();       //     possibly unmark body    
	    X += mass(Bi) * pos(Bi);
	    M += mass(Bi);
	    N0++;
	  }
	iM = M>0.? 1./M : 1.;
	X0 = X*iM;
      }
      // determine c-of-m vel of all unmarked bodies                            
      X = M = 0.;
      LoopBodies(bodies,BB,Bi) if(!is_marked(Bi)) X += mass(Bi) * vel(Bi);
      V0 = X*iM;
      // determine vel dispersion of all unmarked bodies  and                   
      //           number and vel dispersion of all bodies bound to center      
      Nb = 0u;
      S0 = zero;
      Sb = zero;
      register real Mb=zero, vq;
      LoopBodies(bodies,BB,Bi) {                 // loop all bodies           > 
	vq = dist_sq(vel(Bi),V0);                //   compute vq := (v-V0)^2    
	if(!is_marked(Bi))                       //   IF(body unmarked)       > 
	  S0 += mass(Bi) * vq;                   //     add up m * vq         < 
	if(vq + 2*pot(Bi) < zero) {              //   IF(body bound)          > 
	  Mb += mass(Bi);                        //     add up monopole         
	  Sb += mass(Bi) * vq;                   //     add up m * vq           
	  Nb++;                                  //     add up number           
	}                                        //   <                         
      }                                          // <                           
      S0*= iM;                                   // vel disp of unmarked bodies 
      Sb*= Mb>zero? one/Mb : one;                // vel disp of bound bodies    
    }
    //--------------------------------------------------------------------------
    void write(const real& time, ostream& out) const {
      // output of the orbit data                                               
      out<<setprecision(8)<<time  <<" "          // current simulation time     
	 <<setprecision(6)<<X0    <<" "          // center of remaining bodies  
	 <<setprecision(6)<<V0    <<" "          // mean v of remaining bodies  
	 <<setprecision(6)<<S0    <<" "          // v-disp of remaining bodies  
	 <<                 N0    <<" "          // number of remaining bodies  
	 <<setprecision(6)<<Sb    <<" "          // v-disp of bound bodies      
	 <<                 Nb    <<endl;        // number of bound bodies      
    }
  };
}
//------------------------------------------------------------------------------
string defv[] = {
  "in=???\n            file for snapshot input; usually piped from YancNemo  ",
  "out=???\n           file for regular snapshot output                      ",
  "step=1\n            time between output to out                            ",
  "orb=???\n           file for orbit output every time step                 ",
  "r_cut=0\n           radius beyond which bodies are not considered members ",
  "startout=t\n        output first snapshot to file out?                    ",
  "give=mvxy\n         list of output specifications. Recognizing:\n"
  "                     m: mass                          (default)\n"
  "                     x: position                      (default)\n"
  "                     v: velocity                      (default)\n"
  "                     a: acceleration\n"
  "                     p: potential\n"
  "                     y: auxiliary data                (default)          ",
  NULL};
string usage = "AnalysisGC -- Analyis of Globular Cluster simualation\n"
               "              to be used in conjunction with YancNemo";
//------------------------------------------------------------------------------
void nemo_main()
{
  // 1. set some parameters                                                     
  register nbdy::GCOrbit GO(getdparam("r_cut"));
  register nbdy::real    T,t,dt=getdparam("step");
  const    nbdy::io need=nbdy::io::mxvp;
  register nbdy::io write(getparam("give")),
                    want=(write|need) & ~nbdy::io::y;
  register nbdy::bodies BB;
  // 2. initialize I/O streams                                                  
  nbdy::nemo_in  in (getparam("in"));                 // open nemo input stream 
  nbdy::nemo_out out(getparam("out"));                // open nemo output stream
  std::ofstream  orb(getparam("orb"));                // orbit output stream    
  if(!orb.is_open()) error("cannot open orbit file"); // IF(not open): error    
  orb<<"# t x y z vx vy vz sig_r Nr sig_b Nb\n";      // header for orbit output
  // 3. read first input snapshot, initialize aux, & orbit output               
  in.read_history();                                  // read  history          
  out.write_history();                                // write history          
  if(!in.is_present(nbdy::nemo_io::snap))             // IF no snapshot present 
    error("no first snapshot found");                 //   abort                
  in.open_set(nbdy::nemo_io::snap);                   // open nemo snapshot     
  if(want!=BB.read_nemo_particles(in,&T,want,0,true)) //   read data w/o aux    
    error("insufficient data");                       //   abort if too few data
  BB.enable_aux_data();                               //   allocate aux data    
  LoopBodies(nbdy::bodies,(&BB),Bi) Bi.aux() = T-1;   //   set aux data         
  in.close_set(nbdy::nemo_io::snap);                  // close nemo snapshot    
  GO.set(&BB,T);                                      // set orbit              
  GO.write(T,orb);                                    // write orbit            
  if(getbparam("startout"))                           // IF(wanted)           > 
    BB.write_nemo_snapshot(out,&T,write);             //   output snapshot    < 
  T += 0.999999 * dt;                                 // time for snapshot out  
  // 4. loop over input snapshots                                               
  while(in.is_present(nbdy::nemo_io::snap)) {         // WHILE(have snapshots) >
    if(want!=BB.read_nemo_snapshot(in,&t,want,0,true))//   read next snapshot   
      error("insufficient data");                     //   abort if too few data
    GO.set(&BB,t);                                    //   set orbit            
    GO.write(t,orb);                                  //   write orbit          
    if(t >= T) {                                      //   IF(time for output) >
      BB.write_nemo_snapshot(out,&t,write);           //     output snapshot    
      T += dt;                                        //     time for next O    
    }                                                 //   <                    
  }
  // 5. cleaning up (including implicit call of destructors)                    
  in.close();
  out.close();
  orb.close();
}
//---------------------end-of-AnalysisGC.cc------that's-it-!--------------------
