// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// YancNemo.cc                                                                 |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//           YancNemo -- Yet Another N-body Code, for                          |
//                       Not Everybody Must Observe                            |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0.0   28/05/2001  WD created with the help of NEMO caption PJT          |
// v 1.0.1   01/06/2001  WD bug in this file removed, added history output     |
// v 1.0.2   13/06/2001  WD Ncrit option added (before Ncrit==1)               |
// v 1.0.3   20/06/2001  WD tree::make() back to adding-leafs algorithm        |
//                          Coordinatesystem given with snapshot               |
// v 1.0.4   25/06/2001  WD bug (builder::link_tree()) removed (thanks to PJT) |
// v 1.0.5   28/06/2001  WD changes in tree::grow(). Avoid box overflow.       |
//                          bug (in builder) removed                           |
// v 1.0.6   01/08/2001  WD substantial code upgrade in the tree (falcON)      |
// v 1.0.7   02/08/2001  WD Nreuse option added (before Nreuse==0)             |
// v 1.0.8   08/08/2001  WD new data lay-out for bodies, enabling special      |
//                          treatment for leap-frog, saves memory & time (?)   |
// v 1.0.9   17/09/2001  WD added support for dynamically linked external      |
//                          potential (potential.h)                            |
// v 1.0.10  20/09.2001  WD added option ("resume") for resuming an old or     |
//                          interrupted simulation. Appends to input file      |
// v 1.0.11  21/09/2001  WD improved handling of nemo I/O; can now also read   |
//                          PosTag & VelTag instead of PhaseSpaceTag           |
// v 1.0.12  21/09/2001  WD added parameter give_acc                           |
// v 1.0.13  11/10/2001  WD changed give_pot option to allow only N-body pot   |
//                          added option startout                              |
// v 1.0.14  18/10/2001  WD interweaving interaction & evaluation phase of     |
//                          gravity approximation: saves about 20b/body        |
// v 1.0.15  23/10/2001  WD some changes in I/O handling (body, yanc)          |
// v 1.0.16  23/11/2001  WD added option logout (to allow output pipe)         |
// v 1.0.17  08/01/2002  WD changed body lay-out (gives 2% speed-up)           |
// v 1.1.0   21/01/2002  WD minor changes                                      |
// v 1.1.1   25/01/2002  WD minor changes                                      |
// v 1.1.2   29/01/2002  WD minor changes (1.5% speed-up)                      |
// v 1.1.3   26/02/2002  WD time-symmetric stepping criterion tau=fac/acc      |
// v 1.1.4   27/02/2002  WD added options: out2, step2                         |
// v 1.1.5   01/03/2002  WD added time stepping criterion tau=fp/pot           |
// v 1.1.6   05/03/2002  WD added time stepping criterion tau=fc*sqrt(pot)/acc |
// v 1.2     07/06/2002  WD replaced giveacc,givepot,giverho with give, give2  |
// v 1.2.1   11/06/2002  WD support for kernels Fn and Kn withdrawn            |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile YancNemo
#endif

#include <yanc.h>                                     // the N-body code        
#include <nbdy/pext.h>                                // external potential     
using nbdy::io;
using nbdy::yanc;
using nbdy::nemo_pot;

#include <iostream>                                   // C++ I/O                
#include <fstream>                                    // C++ file I/O           
using std::cerr;
using std::ostream;
using std::ofstream;

extern "C" {
#include <nemo.h>                                     // NEMO basics            
#include <potential.h>                                // NEMO potentials        
}
//------------------------------------------------------------------------------
string defv[] = {
  "in=???\n            input file                                         ",
  "out=\n              file for primary output; required, unless resume=t ",
  "tstop=10\n          final integration time                             ",
  "step=1\n            time between primary outputs; 0 -> every step      ",
  "logout=\n           file for log output (default: stderr)              ",
  "out2=\n             file for secondary output stream                   ",
  "step2=0\n           time between secondary outputs; 0 -> every step    ",
  "theta=0.6\n         tolerance parameter at M=M_tot                     ",
  "Nreuse=0\n          max # subsequent tree re-uses                      ",
  "Ncrit=6\n           max # bodies in un-split cells                     ",
  "eps=0.05\n          softening length                                   ",
  "kernel=1\n          softening kernel of family P_n                     ",
  "hmin=6\n            tau_min = (1/2)^hmin                               ",
  "Nlev=1\n            # time-step levels                                 ",
  "fac=\n              tau = fac/acc           \\   If more than one of    ",
  "fph=\n              tau = fph/pot            |  these is non-zero, we  ",
  "fpa=\n              tau = fap*sqrt(pot)/acc /   use the minimum tau.   ",
  "resume=f\n          resume old simulation?  that implies:\n"
  "                   - read last snapshot from input file\n"
  "                   - append primary output to input (unless out given)",
  "give=mxv\n          list of output specifications. Recognizing:\n"
  "                     m: mass                          (default)\n"
  "                     x: position                      (default)\n"
  "                     v: velocity                      (default)\n"
  "                     a: acceleration\n"
  "                     p: N-body potential\n"
  "                     P: external potential\n"
  "                     r: density estimate                              ",
  "give2=mxv\n         list of specifications for secondary output        ",
  "potname=\n          name of external potential                         ",
  "potpars=\n          parameters of external potential                   ",
  "potfile=\n          file required by external potential                ",
  "startout=t\n        primary output for t=tstart?                       ",
  "VERSION=1.2\n       2002 June 7 WD                                     ",
  NULL};
string usage = "YancNemo -- Yet Another N-body Code, for\n"
               "            Not Everybody Must Observe    ";
//------------------------------------------------------------------------------
#if defined(DOUBLE_SINGLE) || defined(DOUBLE_DOUBLE)
# define GET_POT_REAL get_potential_double
#else
# define GET_POT_REAL get_potential_float
#endif
//------------------------------------------------------------------------------
void nemo_main()
{
  // 1. set some parameters                                                     
  register double t_out0,t_out1,                      // t_out0, t_out1         
    t_end   = getdparam("tstop"),                     // integrate until t=t_end
    dt_out0 = getdparam("step"),                      // primary output interval
    dt_out1 = getdparam("step2");                     // secondary  --------    
  register int Nlev = getiparam("Nlev");              // # time step levels     
  if(Nlev>1 &&                                        // IF(more than one level)
     ! (hasvalue("fac") || hasvalue("fph")))          //   we need fac or fph   
    error("fac or fph required if Nlev > 1");         //   error otherwise      
  nemo_pot *pot = hasvalue("potname")?                // IF(potname given) THEN 
    new nemo_pot(GET_POT_REAL(getparam("potname"),    //   initialize external  
			      getparam("potpars"),    //   potential by potname 
			      getparam("potfile")))   //   potpars & potfile    
    : 0;                                              // ELSE: no potential     
  bool resume = getbparam("resume");                  // resume old simulation ?
  io wr0(getparam("give")), wr1(getparam("give2"));   // what to output?        
  // 2. initialize N-body integrator                                            
  yanc YANC(getparam ("in"),                          //   snapshot input       
	    true,                                     //   do nemo I/O          
	    getdparam("theta"),                       //   tolerance parameter  
	    getiparam("Nreuse"),                      //   # re-using tree      
	    getiparam("Ncrit"),                       //   min# bodies/leaf cell
	    getdparam("eps"),                         //   softening length     
	    getiparam("kernel"),                      //   softening kernel     
	    getiparam("hmin"),                        //   -log_2(time step)    
	    Nlev,                                     //   # time-step levels   
	    hasvalue("fac")? getdparam("fac") : 0.,   //   fac in adapting steps
	    hasvalue("fph")? getdparam("fph") : 0.,   //   fph in adapting steps
	    hasvalue("fpa")? getdparam("fpa") : 0.,   //   fpa in adapting steps
	    16u,                                      //   # bodies in rho-estim
	    resume,                                   //   resume old simul?    
	    pot);                                     //   external potential   
  if(!YANC.okay()) error("initialization error\n");   // Initialization okay?   
  if(t_end < YANC.initial_time()) {                   // IF(t_end < t_start)    
    warning("tstop < t_ini: nothing to be done\n");   //   THEN we are done     
    return;                                           //   and can stop here    
  }                                                   // ENDIF                  
  // 3. initialize output streams                                               
  // 3.1 primary output                                                         
  if(hasvalue("out"))                                 // IF(out given) THEN     
    YANC.open_nemo(0,getparam("out"),resume);         //   open NEMO output     
  else if(resume)                                     // ELSE IF(resuming) THEN 
    YANC.open_nemo(0,0,resume);                       //   append to NEMO input 
  else                                                // ELSE(no out, no resume)
    error("out required if resume=f");                //   error out            
  if(!YANC.nemo_is_open(0))                           // IF(no nemo output)     
    error("opening for primary output failed");       //   error out            
  if(!resume && getbparam("startout"))                // IF(not resuming)       
    YANC.write_nemo(wr0,0);                           //   write snapshot       
  // 3.2 secondary output                                                       
  if(hasvalue("out2")) {                              // IF(out2 given) THEN    
    YANC.open_nemo(1,getparam("out2"));               //   try to open output   
      if(!YANC.nemo_is_open(1))                       //   IF(no nemo output)   
	error("opening for secondary output failed"); //     error out          
    YANC.write_nemo(wr1,1);                           //   write snapshot       
  }                                                   // ENDIF                  
  // 3.3 log output                                                             
  ostream *logout;                                    // pointer to log ostream 
  if(hasvalue("logout"))                              // IF desired             
    logout = new ofstream(getparam("logout"));        //   log output to file   
  else                                                // ELSE                   
    logout = &cerr;                                   //   log output to stderr 
  YANC.describe_nemo(*logout,*(ask_history()));       // put history to logout  
  YANC.short_head   (*logout);                        // header for logout      
  YANC.short_stats  (*logout);                        // statistics -> logout   
  // 4. time integration & outputs                                              
  t_out0 = YANC.initial_time() + 0.999999*dt_out0;    // time for next output   
  t_out1 = YANC.initial_time() + 0.999999*dt_out1;    // time for next output   
  while(YANC.time() < t_end) {                        // WHILE( t < t_end )     
    YANC.full_step();                                 //   make full block step 
    YANC.short_stats(*logout);                        //   statistics output    
    if(YANC.time() >= t_out0) {                       //   IF(t >= t_out0)      
      YANC.write_nemo(wr0,0);                         //     primary output     
      t_out0 += dt_out0;                              //     increment t_out0   
    }                                                 //   ENDIF                
    if(YANC.nemo_is_open(1) &&                        //   IF(secondary output  
       YANC.time() >= t_out1) {                       //   AND t >= t_out1)     
      YANC.write_nemo(wr1,1);                         //     secondary output   
      t_out1 += dt_out1;                              //     increment t_out1   
    }                                                 //   ENDIF                
  }                                                   // END: WHILE             
  // 5. cleaning up (including implicit call of destructors)                    
  if(hasvalue("logout"))                              // IF log output to file  
    static_cast<ofstream*>(logout)->close();          //   close log output file
  if(pot) delete pot;                                 // delete external pot    
}
//---------------------end-of-YancNemo.cc------that's-it-!----------------------
