// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// mkcdm.cc                                                                   |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2002                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, 69117 Heidelberg, Germany                         |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// creates N-body initial conditions from a truncated NFW model                |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0    11/06/2002  WD  created for nemo 3.0.12                            |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile mkcdm
#endif

#include <util/Pi.h>
#include <tcdm.h>
#include <body.h>
#include <nbdy/nmio.h>
#include <iomanip>

extern "C" {
#include <nemo.h>
  double xrandom(double,double);
  int    init_xrandom(string);
  double sqrt(double);
  double sin(double);
  double cos(double);
}
////////////////////////////////////////////////////////////////////////////////
string defv[] = {
  "out=???\n         output file                                  ",
  "nbody=???\n       number of bodies                             ",
  "g=???\n           inner density exponent                       ",
  "v=1\n             maximum circular speed                       ",
  "a=1\n             scale radius:      rho_0 = C/r(r+a)^2        ",
  "b=10\n            truncation radius: rho   = rho_0 sech(r/b)   ",
  "seed=0\n          seed for the randum number generator         ",
  "time=0\n          simulation time of snapshot                  ",
  "zerocm=t\n        centrate velocities of snapshot?             ",
  "WD_units=f\n      input:  kpc, km/s\n"
  "                   output: kpc, kpc/Gyr, G=1 (-> mass unit)     ",
  "give_pot=f\n      give potentials with snapshot                ",
  "outputs=f\n       give some global quantities to stderr        ",
  "VERSION=1.0\n     02/April/2002 WD                             ",
  NULL};
string usage =
"mkcdm -- construct a truncated CDM-type halo model with\n\n"
"                         sech(r/b)\n"
"           rho(r) = C ----------------\n"
"                       r^g (r+a)^(3-g)\n\n";
////////////////////////////////////////////////////////////////////////////////
inline double xrandom_() { return xrandom(0.,1.); }
////////////////////////////////////////////////////////////////////////////////
void nemo_main()
{
  //----------------------------------------------------------------------------
  // 1. create truncated NFW model, i.e. integrate to get Phi(r) etc.           
  //----------------------------------------------------------------------------
  const    double vf  = 0.977775320024919, mf = 2.2228847e5;
  register double vcm = getdparam("v");
  if(getbparam("WD_units")) vcm /= vf;
  nbdy::tcdm_model CDM(getdparam("g"), getdparam("a"), getdparam("b"), vcm);
  //----------------------------------------------------------------------------
  // 2. create initial conditions from truncated CDM model                      
  //----------------------------------------------------------------------------
  int N = getiparam("nbody");
  nbdy::bodies BB(N);
  init_xrandom(getparam("seed"));
  const    double     m = CDM.total_mass()/double(N);
  register double     rad,vel,cth,R,phi;
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    //                                                                          
    // 2.1 set mass & potential and get r & v                                   
    //                                                                          
    Bi.mass() = m;
    Bi.pot()  = CDM.random(xrandom_, rad, vel);
    //                                                                          
    // 2.2 get x,y,z                                                            
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = rad * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.pos(0) = R * cos(phi);
    Bi.pos(1) = R * sin(phi);
    Bi.pos(2) = rad * cth;
    //                                                                          
    // 2.3 get vx,vy,vz                                                         
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = vel * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.vel(0) = R * cos(phi);
    Bi.vel(1) = R * sin(phi);
    Bi.vel(2) = vel * cth;
  }
  //----------------------------------------------------------------------------
  // 3. centrate velocity on origin                                             
  //----------------------------------------------------------------------------
  if(getbparam("zerocm")) {
    register nbdy::real iN = nbdy::one/real(N);
    register nbdy::vect cvel(nbdy::zero);
    LoopBodies(nbdy::bodies,(&BB),Bi) cvel += nbdy::vel(Bi);
    cvel *= iN;
    LoopBodies(nbdy::bodies,(&BB),Bi) Bi.vel() -= cvel;
  }
  //----------------------------------------------------------------------------
  // 4. output of snapshot                                                      
  //----------------------------------------------------------------------------
  nbdy::nemo_out out(getparam("out"));
  out.write_history();
  nbdy::real time = getdparam("time");
  BB.write_nemo_snapshot(out,&time,
			 getbparam("give_pot")? nbdy::io::mxv | nbdy::io::p :
			                        nbdy::io::mxv);
  //----------------------------------------------------------------------------
  // 5. optional outputs of global quantities                                   
  //----------------------------------------------------------------------------
  if(getbparam("outputs")) {
    clog<<
      "# ------------------------------------------------------\n"
      "# truncated CDM model:\n"
      "#\n"
      "#                         sech(r/b)\n"
      "#           rho(r) = C ----------------\n"
      "#                       r^g (r+a)^(3-g)\n"
      "#";
    if(getbparam("WD_units"))
      clog<<"\n# C      = "<<std::setw(13)<<CDM.mass_normal()
	  <<" = "<<std::setw(13)<<mf*CDM.mass_normal()   <<" M_sun"
	  <<"\n# M_tot  = "<<std::setw(13)<<CDM.total_mass()
	  <<" = "<<std::setw(13)<<mf*CDM.total_mass()    <<" M_sun"
	  <<"\n# M(r<a) = "<<std::setw(13)<<CDM.cum(CDM.break_radius())
	  <<" = "<<std::setw(13)<<mf*CDM.cum(CDM.break_radius()) <<" M_sun"
	  <<"\n# M(r<b) = "<<std::setw(13)<<CDM.cum(CDM.trunc_radius())
	  <<" = "<<std::setw(13)<<mf*CDM.cum(CDM.trunc_radius()) <<" M_sun"
	  <<"\n# g      = "<<std::setw(13)<<CDM.inner_slope()
	  <<"\n# a      = "<<std::setw(13)<<CDM.break_radius()
	  <<"                 kpc"
	  <<"\n# b      = "<<std::setw(13)<<CDM.trunc_radius()
	  <<"                 kpc"
	  <<"\n# Phi_0  = "<<std::setw(13)<<CDM.central_pot()
	  <<" = "<<std::setw(13)<<vf*vf*CDM.central_pot()<<" (km/s)^2"
	  <<"\n# vc_max = "<<std::setw(13)<<sqrt(CDM.vcirc_square())
	  <<" = "<<std::setw(13)<<vf*sqrt(CDM.vcirc_square())<<" km/s"
	  <<"\n# E_tot  = "<<std::setw(13)<<CDM.total_energy()
	  <<" = "<<std::setw(13)<<vf*vf*mf*CDM.total_energy()<<" M_sun(km/s)^2";
    else
      clog<<"\n# C      = "<<     CDM.mass_normal()
	  <<"\n# M_tot  = "<<     CDM.total_mass()
	  <<"\n# M(r<a) = "<<     CDM.cum(CDM.break_radius())
	  <<"\n# M(r<b) = "<<     CDM.cum(CDM.trunc_radius())
	  <<"\n# a      = "<<     CDM.break_radius()
	  <<"\n# b      = "<<     CDM.trunc_radius()
	  <<"\n# Phi_0  =" <<     CDM.central_pot()
	  <<"\n# vc_max = "<<sqrt(CDM.vcirc_square())
	  <<"\n# E_tot  =" <<     CDM.total_energy();
    clog<<"\n# ------------------------------------------------------"<<endl;
  }
};
